import os
from pathlib import Path
from string import Template
from typing import Tuple, Union, Optional, List

import depthai as dai


class MultiStageConfig:
    debug: bool
    show_cropped_frames: bool
    labels: Optional[List[int]]
    scale_bb: Optional[Tuple[int, int]]

    def __init__(self, debug, labels=None, scale_bb=None):
        self.debug = debug
        self.labels = labels
        self.scale_bb = scale_bb


class MultiStageNN:
    script: dai.node.Script
    manip: dai.node.ImageManip  # Cropping ImageManip
    out: dai.Node.Output  # Cropped imgFrame output
    _size: Tuple[int, int]

    def __init__(self,
                 pipeline: dai.Pipeline,
                 detection_node: Union[dai.node.MobileNetDetectionNetwork,
                                       dai.node.MobileNetSpatialDetectionNetwork,
                                       dai.node.YoloDetectionNetwork,
                                       dai.node.YoloSpatialDetectionNetwork],  # Object detection node
                 high_res_frames: dai.Node.Output,
                 size: Tuple[int, int],
                 debug=False
                 ) -> None:
        """
        Args:

            detections (dai.Node.Output): Object detection output
            high_res_frames (dai.Node.Output): Output that will provide high resolution frames
        """
        # self._input.node, self._input.stream_input, self.size

        self.script = pipeline.create(dai.node.Script)
        self.script.setProcessor(dai.ProcessorType.LEON_CSS)  # More stable
        self._size = size

        detection_node.out.link(self.script.inputs['detections'])
        high_res_frames.link(self.script.inputs['frames'])

        self.configure(MultiStageConfig(debug))

        self.manip = pipeline.create(dai.node.ImageManip)
        self.manip.initialConfig.setResize(size)
        self.manip.setWaitForConfigInput(True)
        self.manip.setMaxOutputFrameSize(size[0] * size[1] * 3)
        self.manip.setNumFramesPool(20)
        self.script.outputs['manip_cfg'].link(self.manip.inputConfig)
        self.script.outputs['manip_img'].link(self.manip.inputImage)
        self.out = self.manip.out

    def configure(self, config: MultiStageConfig = None) -> None:
        """
        Args:
            debug (bool, default False): Debug script node
            labels (List[int], optional): Crop & run inference only on objects with these labels
            scaleBb (Tuple[int, int], optional): Scale detection bounding boxes (x, y) before cropping the frame. In %.
        """
        if config is None:
            return

        with open(Path(os.path.dirname(__file__)) / 'template_multi_stage_script.py', 'r') as file:
            code = Template(file.read()).substitute(
                DEBUG='' if config.debug else '#',
                CHECK_LABELS=f"if det.label not in {str(config.labels)}: continue" if config.labels else "",
                WIDTH=str(self._size[0]),
                HEIGHT=str(self._size[1]),
                SCALE_BB_XMIN=f"-{config.scale_bb[0] / 100}" if config.scale_bb else "",  # % to float value
                SCALE_BB_YMIN=f"-{config.scale_bb[1] / 100}" if config.scale_bb else "",
                SCALE_BB_XMAX=f"+{config.scale_bb[0] / 100}" if config.scale_bb else "",
                SCALE_BB_YMAX=f"+{config.scale_bb[1] / 100}" if config.scale_bb else "",
            )
            self.script.setScript(code)
            # print(f"\n------------{code}\n---------------")



from pathlib import Path
from typing import Any, Generator, List, Dict, Tuple

import cv2
import numpy as np
from rosbags.rosbag2 import Reader
from rosbags.serde import deserialize_cdr

from depthai_sdk.previews import PreviewDecoder
from depthai_sdk.readers.abstract_reader import AbstractReader


class Db3Reader(AbstractReader):
    STREAMS = ['left', 'right', 'depth']
    generators: Dict[str, Generator] = {}
    frames = None  # For shapes

    def __init__(self, folder: Path) -> None:
        self.reader = Reader(str(folder))
        self.reader.open()

        for con in self.reader.connections:
            for stream in self.STREAMS:
                if stream.lower() in con.topic.lower():
                    self.generators[stream.lower()] = self.reader.messages([con])

        # For shapes
        self.frames = self.read()

        for con in self.reader.connections:
            for stream in self.STREAMS:
                if stream.lower() in con.topic.lower():
                    self.generators[stream.lower()] = self.reader.messages([con])

    def read(self):
        rosMsgs: Dict[str, Any] = dict()
        try:
            for name, gen in self.generators.items():
                con, ts, raw = next(gen)
                rosMsgs[name] = self._getCvFrame(deserialize_cdr(raw, con.msgtype), name)
            return rosMsgs
        except:
            return None

    def disableStream(self, name: str):
        if name in self.generators:
            del self.generators[name]

    def _getCvFrame(self, msg, name: str):
        """
        Convert ROS message to cv2 frame (numpy array)
        """
        msgType = str(type(msg))
        data = np.frombuffer(msg.data, dtype=np.int8)
        if 'CompressedImage' in msgType:
            if name == 'color':
                return PreviewDecoder.jpegDecode(data, cv2.IMREAD_COLOR)
            else:  # left, right, disparity
                return PreviewDecoder.jpegDecode(data, cv2.IMREAD_GRAYSCALE)
        elif 'Image' in msgType:
            if msg.encoding == 'mono16':
                data = data.view(np.int16)
            return data.reshape((msg.height, msg.width))
            # msg.encoding
        else:
            raise Exception('Only CompressedImage and Image ROS messages are currently supported.')

    def getStreams(self) -> List[str]:
        streams = [name for name in self.frames]
        return streams

    def getShape(self, name: str) -> Tuple[int, int]:
        frame = self.frames[name]
        return (frame.shape[1], frame.shape[0])

    def close(self):
        self.reader.close()



from datetime import timedelta
from fractions import Fraction
from pathlib import Path

import depthai as dai

from depthai_sdk.recorders.video_writers import AbstractWriter
from depthai_sdk.recorders.video_writers.utils import create_writer_dir


class AvWriter(AbstractWriter):
    start_ts: timedelta = None
    file = None

    def __init__(self, folder: Path, name: str, fourcc: str, fps: float):
        global av
        import av as av

        name = create_writer_dir(folder, name, 'mp4')

        self.start_ts = None
        self.folder = folder
        self.file = av.open(name, 'w')
        self._fps = fps
        self._fourcc = fourcc

        self._create_stream(self.file, fourcc, fps)

        # self._buffer = None
        # if self._keep_last_seconds > 0:
        #     self._buffer = deque(maxlen=int(10 * fps))

    def _create_stream(self, file, fourcc, fps) -> None:
        """Create stream in file with given fourcc and fps, works in-place."""
        stream = file.add_stream(fourcc, rate=int(fps))
        stream.time_base = Fraction(1, 1000 * 1000)  # Microseconds

        # We need to set pixel format for MJEPG, for H264/H265 it's yuv420p by default
        if fourcc == 'mjpeg':
            stream.pix_fmt = 'yuvj420p'

    def write(self, frame: dai.ImgFrame) -> None:
        packet = av.Packet(frame.getData())  # Create new packet with byte array

        # if self._buffer is not None:
        #     if len(self._buffer) == self._buffer.maxlen:
        #         self._buffer.pop()
        #
        #     self._buffer.append(frame)

        # Set frame timestamp
        if self.start_ts is None:
            self.start_ts = frame.getTimestampDevice()

        ts = int((frame.getTimestampDevice() - self.start_ts).total_seconds() * 1e6)  # To microsec
        packet.dts = ts
        packet.pts = ts

        self.file.mux_one(packet)  # Mux the Packet into container

    # def get_last(self, seconds: float = 0.0):
    #     if self._buffer is None:
    #         print('Buffer is not initialized, cannot get last frames')
    #         return
    #
    #     if len(self._buffer) == 0:
    #         print('Buffer is empty, cannot get last frames')
    #         return
    #
    #     snapshot_path = f'{self.folder}/snapshot_{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}.mp4'
    #     snapshot_file = av.open(snapshot_path, 'w')
    #     self._create_stream(snapshot_file, self._fourcc, self._fps)
    #
    #     # Copy queue
    #     buffer_copy = self._buffer.copy()
    #
    #     start_ts = None
    #     n_skip_frames = int(self._fps * (self._keep_last_seconds - seconds))
    #     while len(buffer_copy) > 0:
    #         # Wait til we reach the desired time
    #         if n_skip_frames > 0:
    #             n_skip_frames -= 1
    #             buffer_copy.popleft()
    #             continue
    #
    #         el = buffer_copy.popleft()
    #         packet = av.Packet(el.getData())
    #         if start_ts is None:
    #             start_ts = el.getTimestampDevice()
    #
    #         ts = int((el.getTimestampDevice() - start_ts).total_seconds() * 1e6)  # To microsec
    #         packet.dts = ts
    #         packet.pts = ts
    #
    #         snapshot_file.mux(packet)
    #
    #     snapshot_file.close()
    #     print('Snapshot saved to', snapshot_path)

    def close(self):
        self.file.close()



import cv2
import numpy as np

from depthai_sdk import toTensorResult, Previews

keypointsMapping = ['Nose', 'Neck', 'R-Sho', 'R-Elb', 'R-Wr', 'L-Sho', 'L-Elb', 'L-Wr', 'R-Hip', 'R-Knee', 'R-Ank',
                    'L-Hip', 'L-Knee', 'L-Ank', 'R-Eye', 'L-Eye', 'R-Ear', 'L-Ear']
POSE_PAIRS = [[1, 2], [1, 5], [2, 3], [3, 4], [5, 6], [6, 7], [1, 8], [8, 9], [9, 10], [1, 11], [11, 12], [12, 13],
              [1, 0], [0, 14], [14, 16], [0, 15], [15, 17], [2, 17], [5, 16]]
mapIdx = [[31, 32], [39, 40], [33, 34], [35, 36], [41, 42], [43, 44], [19, 20], [21, 22], [23, 24], [25, 26], [27, 28],
          [29, 30], [47, 48], [49, 50], [53, 54], [51, 52], [55, 56], [37, 38], [45, 46]]
colors = [[0, 100, 255], [0, 100, 255], [0, 255, 255], [0, 100, 255], [0, 255, 255], [0, 100, 255], [0, 255, 0],
          [255, 200, 100], [255, 0, 255], [0, 255, 0], [255, 200, 100], [255, 0, 255], [0, 0, 255], [255, 0, 0],
          [200, 200, 0], [255, 0, 0], [200, 200, 0], [0, 0, 0]]


def getKeypoints(probMap, threshold=0.2):
    mapSmooth = cv2.GaussianBlur(probMap, (3, 3), 0, 0)
    mapMask = np.uint8(mapSmooth > threshold)
    keypoints = []
    contours = None
    try:
        # OpenCV4.x
        contours, _ = cv2.findContours(mapMask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    except:
        # OpenCV3.x
        _, contours, _ = cv2.findContours(mapMask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    for cnt in contours:
        blobMask = np.zeros(mapMask.shape)
        blobMask = cv2.fillConvexPoly(blobMask, cnt, 1)
        maskedProbMap = mapSmooth * blobMask
        _, maxVal, _, maxLoc = cv2.minMaxLoc(maskedProbMap)
        keypoints.append(maxLoc + (probMap[maxLoc[1], maxLoc[0]],))

    return keypoints


def getValidPairs(outputs, w, h, detectedKeypoints):
    validPairs = []
    invalidPairs = []
    nInterpSamples = 10
    pafScoreTh = 0.2
    confTh = 0.4
    for k in range(len(mapIdx)):

        pafA = outputs[0, mapIdx[k][0], :, :]
        pafB = outputs[0, mapIdx[k][1], :, :]
        pafA = cv2.resize(pafA, (w, h))
        pafB = cv2.resize(pafB, (w, h))
        candA = detectedKeypoints[POSE_PAIRS[k][0]]

        candB = detectedKeypoints[POSE_PAIRS[k][1]]

        nA = len(candA)
        nB = len(candB)

        if (nA != 0 and nB != 0):
            validPair = np.zeros((0, 3))
            for i in range(nA):
                maxJ = -1
                maxScore = -1
                found = 0
                for j in range(nB):
                    d_ij = np.subtract(candB[j][:2], candA[i][:2])
                    norm = np.linalg.norm(d_ij)
                    if norm:
                        d_ij = d_ij / norm
                    else:
                        continue
                    interp_coord = list(zip(np.linspace(candA[i][0], candB[j][0], num=nInterpSamples),
                                            np.linspace(candA[i][1], candB[j][1], num=nInterpSamples)))
                    pafInterp = []
                    for k in range(len(interp_coord)):
                        pafInterp.append([pafA[int(round(interp_coord[k][1])), int(round(interp_coord[k][0]))],
                                           pafB[int(round(interp_coord[k][1])), int(round(interp_coord[k][0]))]])
                    pafScores = np.dot(pafInterp, d_ij)
                    avgPafScore = sum(pafScores) / len(pafScores)

                    if (len(np.where(pafScores > pafScoreTh)[0]) / nInterpSamples) > confTh:
                        if avgPafScore > maxScore:
                            maxJ = j
                            maxScore = avgPafScore
                            found = 1
                if found:
                    validPair = np.append(validPair, [[candA[i][3], candB[maxJ][3], maxScore]], axis=0)

            validPairs.append(validPair)
        else:
            invalidPairs.append(k)
            validPairs.append([])
    return validPairs, invalidPairs


def getPersonwiseKeypoints(validPairs, invalidPairs, keypointsList):
    personwiseKeypoints = -1 * np.ones((0, 19))

    for k in range(len(mapIdx)):
        if k not in invalidPairs:
            partAs = validPairs[k][:, 0]
            partBs = validPairs[k][:, 1]
            indexA, indexB = np.array(POSE_PAIRS[k])

            for i in range(len(validPairs[k])):
                found = 0
                personIdx = -1
                for j in range(len(personwiseKeypoints)):
                    if personwiseKeypoints[j][indexA] == partAs[i]:
                        personIdx = j
                        found = 1
                        break

                if found:
                    personwiseKeypoints[personIdx][indexB] = partBs[i]
                    personwiseKeypoints[personIdx][-1] += keypointsList[partBs[i].astype(int), 2] + validPairs[k][i][
                        2]

                elif not found and k < 17:
                    row = -1 * np.ones(19)
                    row[indexA] = partAs[i]
                    row[indexB] = partBs[i]
                    row[-1] = sum(keypointsList[validPairs[k][i, :2].astype(int), 2]) + validPairs[k][i][2]
                    personwiseKeypoints = np.vstack([personwiseKeypoints, row])
    return personwiseKeypoints


threshold = 0.3
nPoints = 18
detectedKeypoints = []


def decode(nnManager, packet):
    outputs = toTensorResult(packet)["Openpose/concat_stage7"].astype('float32')
    w, h = nnManager.inputSize

    detectedKeypoints = []
    keypointsList = np.zeros((0, 3))
    keypointId = 0

    for part in range(nPoints):
        probMap = outputs[0, part, :, :]
        probMap = cv2.resize(probMap, (w, h))  # (456, 256)
        keypoints = getKeypoints(probMap, threshold)
        keypointsWithId = []

        for i in range(len(keypoints)):
            keypointsWithId.append(keypoints[i] + (keypointId,))
            keypointsList = np.vstack([keypointsList, keypoints[i]])
            keypointId += 1

        detectedKeypoints.append(keypointsWithId)

    validPairs, invalidPairs = getValidPairs(outputs, w, h, detectedKeypoints)
    personwiseKeypoints = getPersonwiseKeypoints(validPairs, invalidPairs, keypointsList)
    keypointsLimbs = [detectedKeypoints, personwiseKeypoints, keypointsList]

    return keypointsLimbs


def draw(nnManager, keypointsLimbs, frames):
    for name, frame in frames:
        if name == "color" and nnManager.source == "color" and not nnManager._fullFov:
            scaleFactor = frame.shape[0] / nnManager.inputSize[1]
            offsetW = int(frame.shape[1] - nnManager.inputSize[0] * scaleFactor) // 2

            def scale(point):
                return int(point[0] * scaleFactor) + offsetW, int(point[1] * scaleFactor)
        elif name in (Previews.color.name, Previews.nnInput.name, "host"):
            scaleH = frame.shape[0] / nnManager.inputSize[1]
            scaleW = frame.shape[1] / nnManager.inputSize[0]

            def scale(point):
                return int(point[0] * scaleW), int(point[1] * scaleH)
        else:
            continue

        if len(keypointsLimbs) == 3:
            detectedKeypoints = keypointsLimbs[0]
            personwiseKeypoints = keypointsLimbs[1]
            keypointsList = keypointsLimbs[2]

            for i in range(nPoints):
                for j in range(len(detectedKeypoints[i])):
                    cv2.circle(frame, scale(detectedKeypoints[i][j][0:2]), 5, colors[i], -1, cv2.LINE_AA)
            for i in range(17):
                for n in range(len(personwiseKeypoints)):
                    index = personwiseKeypoints[n][np.array(POSE_PAIRS[i])]
                    if -1 in index:
                        continue
                    B = np.int32(keypointsList[index.astype(int), 0])
                    A = np.int32(keypointsList[index.astype(int), 1])
                    cv2.line(frame, scale((B[0], A[0])), scale((B[1], A[1])), colors[i], 3, cv2.LINE_AA)




import os
import time
from pathlib import Path
from typing import List, Tuple, Dict

import cv2
import numpy as np

from depthai_sdk.readers.abstract_reader import AbstractReader

# Supported image formats:
_imageExt = ['.bmp', '.dib', '.jpeg', '.jpg', '.jpe', '.jp2', '.png', '.webp', '.pbm', '.pgm', '.ppm', '.pxm',
             '.pnm', '.pfm', '.sr', '.ras', '.tiff', '.tif', '.exr', '.hdr', '.pic']


def get_name_flag(name: str) -> Tuple:
    stream = 'color'
    flag = cv2.IMREAD_COLOR
    if name == 'left' or name == 'right':
        flag = cv2.IMREAD_GRAYSCALE
        stream = name
    return (stream, flag)


class ImageReader(AbstractReader):
    """
    Reads image(s) in the file. If file name is 'left'/'right', it will take those as the stereo camera pair, otherwise
    it will send them as color frames. If it has multiple frames of the same stream it will cycle them with 1.5 sec
    delay.
    """
    frames: Dict[str, List] = {
        'color': [],
        'left': [],
        'right': [],
    }
    cntr: Dict[str, int] = dict()

    def __init__(self, path: Path) -> None:
        if path.is_file():
            stream, flag = get_name_flag(path.stem)
            self.frames[stream].append(cv2.imread(str(path), flag))
        else:
            shape = None
            for fileName in os.listdir(str(path)):
                f_name, ext = os.path.splitext(fileName)
                if ext not in _imageExt: continue
                stream, flag = get_name_flag(f_name)

                frame = cv2.imread(str(path / fileName), flag)
                if shape is None:
                    shape = (frame.shape[1], frame.shape[0])
                # Resize, so all images are the same size
                frame = cv2.resize(frame, shape)

                self.frames[stream].append(frame)

        for name, arr in self.frames.items():
            self.cntr[name] = 0

        self.cycle = time.time()

    def read(self):
        # Increase counters
        if 3 < time.time() - self.cycle:
            self.cycle = time.time()
            for name in self.cntr:
                self.cntr[name] += 1
                if len(self.frames[name]) <= self.cntr[name]:
                    self.cntr[name] = 0

        msgs: Dict[str, np.ndarray] = dict()
        for name, arr in self.frames.items():
            if 0 < len(arr):
                msgs[name] = arr[self.cntr[name]]
        return msgs

    def getStreams(self) -> List[str]:
        streams = []
        for name, arr in self.frames.items():
            if 0 < len(arr):
                streams.append(name)
        return streams

    def getShape(self, name: str) -> Tuple[int, int]:
        shape = self.frames[name][0].shape
        return (shape[1], shape[0])

    def close(self):
        pass

    def disableStream(self, name: str):
        if name in self.frames:
            self.frames[name] = []



#!/usr/bin/python
import json
from collections import defaultdict

from ..constants import (
    NODE_PROP,
    NODE_PROP_QLABEL,
    NODE_PROP_QLINEEDIT,
    NODE_PROP_QCHECKBOX,
    NODE_PROP_COLORPICKER
)
from ..errors import NodePropertyError


class PortModel(object):
    """
    Data dump for a port object.
    """

    def __init__(self, node):
        self.node = node
        self.type_ = ''
        self.name = 'port'
        self.display_name = True
        self.multi_connection = False
        self.visible = True
        self.locked = False
        self.connected_ports = defaultdict(list)

    def __repr__(self):
        return '<{}(\'{}\') object at {}>'.format(
            self.__class__.__name__, self.name, hex(id(self)))

    @property
    def to_dict(self):
        """
        serialize model information to a dictionary.

        Returns:
            dict: node port dictionary eg.
                {
                    'type': 'in',
                    'name': 'port',
                    'display_name': True,
                    'multi_connection': False,
                    'visible': True,
                    'locked': False,
                    'connected_ports': {<node_id>: [<port_name>, <port_name>]}
                }
        """
        props = self.__dict__.copy()
        props.pop('node')
        props['connected_ports'] = dict(props.pop('connected_ports'))
        return props


class NodeModel(object):
    """
    Data dump for a node object.
    """

    def __init__(self):
        self.type_ = None
        self.id = hex(id(self))
        self.icon = None
        self.name = 'node'
        self.color = (13, 18, 23, 255)
        self.border_color = (74, 84, 85, 255)
        self.text_color = (255, 255, 255, 180)
        self.disabled = False
        self.selected = False
        self.visible = True
        self.width = 100.0
        self.height = 80.0
        self.pos = [0.0, 0.0]

        # BaseNode attrs.
        self.inputs = {}
        self.outputs = {}

        # To save ports in json file
        # self.port_deletion_allowed = False
        self.port_deletion_allowed = True

        # GroupNode attrs.
        self.subgraph_session = {}

        # Custom
        self._custom_prop = {}

        # node graph model set at node added time.
        self._graph_model = None

        # store the property attributes.
        # (deleted when node is added to the graph)
        self._TEMP_property_attrs = {}

        # temp store the property widget types.
        # (deleted when node is added to the graph)
        self._TEMP_property_widget_types = {
            'type_': NODE_PROP_QLABEL,
            'id': NODE_PROP_QLABEL,
            'icon': NODE_PROP,
            'name': NODE_PROP_QLINEEDIT,
            'color': NODE_PROP_COLORPICKER,
            'border_color': NODE_PROP,
            'text_color': NODE_PROP_COLORPICKER,
            'disabled': NODE_PROP_QCHECKBOX,
            'selected': NODE_PROP,
            'width': NODE_PROP,
            'height': NODE_PROP,
            'pos': NODE_PROP,
            'inputs': NODE_PROP,
            'outputs': NODE_PROP,
        }

    def __repr__(self):
        return '<{}(\'{}\') object at {}>'.format(
            self.__class__.__name__, self.name, self.id)

    def add_property(self, name, value, items=None, range=None,
                     widget_type=NODE_PROP, tab=None):
        """
        add custom property.

        Args:
            name (str): name of the property.
            value (object): data.
            items (list[str]): items used by widget type NODE_PROP_QCOMBO.
            range (tuple)): min, max values used by NODE_PROP_SLIDER.
            widget_type (int): widget type flag.
            tab (str): widget tab name.
        """
        tab = tab or 'Properties'

        if name in self.properties.keys():
            raise NodePropertyError(
                '"{}" reserved for default property.'.format(name))
        if name in self._custom_prop.keys():
            raise NodePropertyError(
                '"{}" property already exists.'.format(name))

        self._custom_prop[name] = value

        if self._graph_model is None:
            self._TEMP_property_widget_types[name] = widget_type
            self._TEMP_property_attrs[name] = {'tab': tab}
            if items:
                self._TEMP_property_attrs[name]['items'] = items
            if range:
                self._TEMP_property_attrs[name]['range'] = range
        else:
            attrs = {self.type_: {name: {
                'widget_type': widget_type,
                'tab': tab
            }}}
            if items:
                attrs[self.type_][name]['items'] = items
            if range:
                attrs[self.type_][name]['range'] = range
            self._graph_model.set_node_common_properties(attrs)

    def set_property(self, name, value):
        if name in self.properties.keys():
            setattr(self, name, value)
        elif name in self._custom_prop.keys():
            self._custom_prop[name] = value
        else:
            raise NodePropertyError('No property "{}"'.format(name))

    def get_property(self, name):
        if name in self.properties.keys():
            return self.properties[name]
        return self._custom_prop.get(name)

    def get_widget_type(self, name):
        model = self._graph_model
        if model is None:
            return self._TEMP_property_widget_types.get(name)
        return model.get_node_common_properties(self.type_)[name]['widget_type']

    def get_tab_name(self, name):
        model = self._graph_model
        if model is None:
            attrs = self._TEMP_property_attrs.get(name)
            if attrs:
                return attrs[name].get('tab')
            return
        return model.get_node_common_properties(self.type_)[name]['tab']

    @property
    def properties(self):
        """
        return all default node properties.

        Returns:
            dict: default node properties.
        """
        props = self.__dict__.copy()
        exclude = ['_custom_prop',
                   '_graph_model',
                   '_TEMP_property_attrs',
                   '_TEMP_property_widget_types']
        [props.pop(i) for i in exclude if i in props.keys()]
        return props

    @property
    def custom_properties(self):
        """
        return all custom properties specified by the user.

        Returns:
            dict: user defined properties.
        """
        return self._custom_prop

    @property
    def to_dict(self):
        """
        serialize model information to a dictionary.

        Returns:
            dict: node id as the key and properties as the values eg.
                {'0x106cf75a8': {
                    'name': 'foo node',
                    'color': (48, 58, 69, 255),
                    'border_color': (85, 100, 100, 255),
                    'text_color': (255, 255, 255, 180),
                    'type_': 'com.chantasticvfx.FooNode',
                    'selected': False,
                    'disabled': False,
                    'visible': True,
                    'width': 0.0,
                    'height: 0.0,
                    'pos': (0.0, 0.0),
                    'custom': {},
                    'inputs': {
                        <port_name>: {<node_id>: [<port_name>, <port_name>]}
                    },
                    'outputs': {
                        <port_name>: {<node_id>: [<port_name>, <port_name>]}
                    },
                    'input_ports': [<port_name>, <port_name>],
                    'output_ports': [<port_name>, <port_name>],
                    },
                    subgraph_session: <sub graph session data>
                }
        """
        node_dict = self.__dict__.copy()
        node_id = node_dict.pop('id')

        inputs = {}
        outputs = {}
        input_ports = []
        output_ports = []
        for name, model in node_dict.pop('inputs').items():
            if self.port_deletion_allowed:
                # GX Find the port color
                for inp in model.node._inputs:
                    if inp._Port__model is model:
                        port_color = inp._Port__view._color
                        break
                input_ports.append({
                    'name': name,
                    'multi_connection': model.multi_connection,
                    'display_name': model.display_name,
                    'color': port_color
                })
            connected_ports = model.to_dict['connected_ports']
            if connected_ports:
                inputs[name] = connected_ports
        for name, model in node_dict.pop('outputs').items():
            if self.port_deletion_allowed:
                output_ports.append({
                    'name': name,
                    'multi_connection': model.multi_connection,
                    'display_name': model.display_name,
                })
            connected_ports = model.to_dict['connected_ports']
            if connected_ports:
                outputs[name] = connected_ports
        if inputs:
            node_dict['inputs'] = inputs
        if outputs:
            node_dict['outputs'] = outputs

        if self.port_deletion_allowed:
            node_dict['input_ports'] = input_ports
            node_dict['output_ports'] = output_ports

        if self.subgraph_session:
            node_dict['subgraph_session'] = self.subgraph_session

        custom_props = node_dict.pop('_custom_prop', {})
        if custom_props:
            node_dict['custom'] = custom_props

        exclude = ['_graph_model',
                   '_TEMP_property_attrs',
                   '_TEMP_property_widget_types']
        [node_dict.pop(i) for i in exclude if i in node_dict.keys()]

        return {node_id: node_dict}

    @property
    def serial(self):
        """
        Serialize model information to a string.

        Returns:
            str: serialized JSON string.
        """
        model_dict = self.to_dict
        return json.dumps(model_dict)


class NodeGraphModel(object):
    """
    Data dump for a node graph.
    """

    def __init__(self):
        self.__common_node_props = {}

        self.nodes = {}
        self.session = ''
        self.acyclic = False
        self.pipe_collision = False

    def common_properties(self):
        """
        Return all common node properties.

        Returns:
            dict: common node properties.
                eg.
                    {'nodeGraphQt.nodes.FooNode': {
                        'my_property': {
                            'widget_type': 0,
                            'tab': 'Properties',
                            'items': ['foo', 'bar', 'test'],
                            'range': (0, 100)
                            }
                        }
                    }
        """
        return self.__common_node_props

    def set_node_common_properties(self, attrs):
        """
        Store common node properties.

        Args:
            attrs (dict): common node properties.
                eg.
                    {'nodeGraphQt.nodes.FooNode': {
                        'my_property': {
                            'widget_type': 0,
                            'tab': 'Properties',
                            'items': ['foo', 'bar', 'test'],
                            'range': (0, 100)
                            }
                        }
                    }
        """
        for node_type in attrs.keys():
            node_props = attrs[node_type]

            if node_type not in self.__common_node_props.keys():
                self.__common_node_props[node_type] = node_props
                continue

            for prop_name, prop_attrs in node_props.items():
                common_props = self.__common_node_props[node_type]
                if prop_name not in common_props.keys():
                    common_props[prop_name] = prop_attrs
                    continue
                common_props[prop_name].update(prop_attrs)

    def get_node_common_properties(self, node_type):
        """
        Return all the common properties for a registered node.

        Args:
            node_type (str): node type.

        Returns:
            dict: node common properties.
        """
        return self.__common_node_props.get(node_type)


if __name__ == '__main__':
    p = PortModel(None)
    # print(p.to_dict)

    n = NodeModel()
    n.inputs[p.name] = p
    n.add_property('foo', 'bar')

    print('-'*100)
    print('property keys\n')
    print(list(n.properties.keys()))
    print('-'*100)
    print('to_dict\n')
    for k, v in n.to_dict[n.id].items():
        print(k, v)



#!/usr/bin/python
from ..constants import (NODE_LAYOUT_VERTICAL,
                                   NODE_LAYOUT_HORIZONTAL)

from ..errors import PortRegistrationError
from ..nodes.base_node import BaseNode
from ..qgraphics.node_port_in import (PortInputNodeItem,
                                                PortInputNodeVerticalItem)
from ..qgraphics.node_port_out import (PortOutputNodeItem,
                                                 PortOutputNodeVerticalItem)


class PortInputNode(BaseNode):
    """
    The ``PortInputNode`` class is the node object that represents a port from a
    :class:`NodeGraphQt.GroupNode` when expanded in a
    :class:`NodeGraphQt.SubGraph`.

    **Inherited from:** :class:`NodeGraphQt.BaseNode`

    .. image:: ../_images/port_in_node.png
        :width: 150px

    -
    """

    NODE_NAME = 'InputPort'

    def __init__(self, qgraphics_views=None, parent_port=None):
        qgraphics_views = qgraphics_views or {
            NODE_LAYOUT_HORIZONTAL: PortInputNodeItem,
            NODE_LAYOUT_VERTICAL: PortInputNodeVerticalItem
        }
        super(PortInputNode, self).__init__(qgraphics_views)
        self._parent_port = parent_port

    @property
    def parent_port(self):
        """
        The parent group node port representing this node.

        Returns:
            NodeGraphQt.Port: port object.
        """
        return self._parent_port

    def add_input(self, name='input', multi_input=False, display_name=True,
                  color=None, locked=False, painter_func=None):
        """
        This is not available for the `PortInputNode` class.
        """
        raise PortRegistrationError(
            '"{}.add_input()" is not available for {}.'
            .format(self.__class__.__name__, self)
        )

    def add_output(self, name='output', multi_output=True, display_name=True,
                   color=None, locked=False, painter_func=None):
        if self._outputs:
            raise PortRegistrationError(
                '"{}.add_output()" only ONE output is allowed for this node.'
                .format(self.__class__.__name__, self)
            )
        super(PortInputNode, self).add_output(
            name=name,
            multi_output=multi_output,
            display_name=False,
            color=color,
            locked=locked,
            painter_func=None
        )


class PortOutputNode(BaseNode):
    """
    The ``PortOutputNode`` class is the node object that represents a port from a
    :class:`NodeGraphQt.GroupNode` when expanded in a
    :class:`NodeGraphQt.SubGraph`.

    **Inherited from:** :class:`NodeGraphQt.BaseNode`

    .. image:: ../_images/port_out_node.png
        :width: 150px

    -
    """

    NODE_NAME = 'OutputPort'

    def __init__(self, qgraphics_views=None, parent_port=None):
        qgraphics_views = qgraphics_views or {
            NODE_LAYOUT_HORIZONTAL: PortOutputNodeItem,
            NODE_LAYOUT_VERTICAL: PortOutputNodeVerticalItem
        }
        super(PortOutputNode, self).__init__(qgraphics_views)
        self._parent_port = parent_port

    @property
    def parent_port(self):
        """
        The parent group node port representing this node.

        Returns:
            NodeGraphQt.Port: port object.
        """
        return self._parent_port

    def add_input(self, name='input', multi_input=False, display_name=True,
                  color=None, locked=False, painter_func=None):
        if self._inputs:
            raise PortRegistrationError(
                '"{}.add_input()" only ONE input is allowed for this node.'
                .format(self.__class__.__name__, self)
            )
        super(PortOutputNode, self).add_input(
            name=name,
            multi_input=multi_input,
            display_name=False,
            color=color,
            locked=locked,
            painter_func=None
        )

    def add_output(self, name='output', multi_output=True, display_name=True,
                   color=None, locked=False, painter_func=None):
        """
        This is not available for the `PortOutputNode` class.
        """
        raise PortRegistrationError(
            '"{}.add_output()" is not available for {}.'
            .format(self.__class__.__name__, self)
        )



from depthai_sdk.args_parser import ArgsParser
from depthai_sdk.classes import *
from depthai_sdk.classes.packets import *
from depthai_sdk.components import *
from depthai_sdk.components.nn_helper import AspectRatioResizeMode
from depthai_sdk.fps import *
from depthai_sdk.oak_camera import OakCamera
from depthai_sdk.oak_device import OakDevice
from depthai_sdk.oak_outputs.fps import FPS
from depthai_sdk.previews import *
from depthai_sdk.record import *
from depthai_sdk.replay import *
from depthai_sdk.utils import *
from depthai_sdk.visualize import *
from depthai_sdk.visualize.visualizer_helper import VisualizerHelper



#!/usr/bin/env python3
from enum import IntEnum
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Dict, List

import depthai as dai

from depthai_sdk.classes.packets import FramePacket
from depthai_sdk.oak_outputs.xout import XoutSeqSync, XoutFrames
from depthai_sdk.recorders.abstract_recorder import Recorder


def _run(recorder: Recorder, frameQ: Queue):
    """
    Start recording infinite loop
    """
    while True:
        try:
            frames = frameQ.get()
            if frames is None:  # Terminate app
                break

            for name in frames:
                # Save all synced frames into files
                recorder.write(name, frames[name])
        except KeyboardInterrupt:
            break
    # Close all recorders - Can't use ExitStack with VideoWriter
    recorder.close()
    print('Exiting store frame thread')


class RecordType(IntEnum):
    VIDEO = 1  # Save to video file
    BAG = 2  # To ROS .bag
    MCAP = 3  # To .mcap


class Record(XoutSeqSync):
    """
    This class records depthai streams from OAK cameras into different formats.
    Available formats: .h265, .mjpeg, .mp4, .mcap, .bag
    It will also save calibration .json, so depth reconstruction will
    """
    name_mapping: Dict[str, str]  # XLinkOut stream name -> Friendly name mapping
    frame_q: Queue = None
    recorder: Recorder

    def package(self, msgs: Dict):
        # Here we get sequence-num synced messages:)
        mapped = dict()
        for name, msg in msgs.items():
            if name in self.name_mapping:  # Map to friendly name
                mapped[self.name_mapping[name]] = msg
            else:
                mapped[name] = msg

        self.frame_q.put(mapped)

    def visualize(self, packet: FramePacket) -> None:
        pass  # No need.

    def __init__(self,
                 path: Path,
                 type: RecordType):
        """
        Args:
            path (Path): Path to the recording folder
            device (dai.Device): OAK device object
        """
        super().__init__([])  # We don't yet have streams, we will set it up later
        self.folder = path
        self.type = type

        if self.type == RecordType.MCAP:
            from .recorders.mcap_recorder import McapRecorder
            self.recorder = McapRecorder()
        elif self.type == RecordType.VIDEO:
            from .recorders.video_recorder import VideoRecorder
            self.recorder = VideoRecorder()
        elif self.type == RecordType.BAG:
            from .recorders.rosbag_recorder import RosbagRecorder
            self.recorder = RosbagRecorder()
        else:
            raise ValueError(f"Recording type '{self.type}' isn't supported!")

    def no_sync(self, name: str, msg):
        # name = self.name_mapping[name] if name in self.name_mapping else name
        obj = {name: msg}
        self.frame_q.put(obj)

    def start(self, device: dai.Device, xouts: List[XoutFrames]):
        """
        Start recording process. This will create and start the pipeline,
        start recording threads, and initialize all queues.
        """
        if self.type == RecordType.VIDEO:
            self._streams = [out.frames.name for out in xouts]  # required by XoutSeqSync
            self.streamNum = len(xouts)
            self.name_mapping = dict()
            for xout in xouts:
                self.name_mapping[xout.frames.name] = xout.name
        else: # For MCAP/Ros bags we don't need msg syncing
            self.newMsg = self.no_sync

        self.mxid = device.getMxId()
        self.path = self._createFolder(self.folder, self.mxid)
        calibData = device.readCalibration()
        calibData.eepromToJsonFile(str(self.path / "calib.json"))

        self.recorder.update(self.path, device, xouts)

        self.frame_q = Queue(maxsize=20)
        self.process = Thread(target=_run, args=(self.recorder, self.frame_q))
        self.process.start()

    # TODO: support pointclouds in MCAP
    def config_mcap(self, pointcloud: bool):
        if self.type != RecordType.MCAP:
            print(f"Recorder type is {self.type}, not MCAP! Config attempt ignored.")
            return
        self.recorder.setPointcloud(pointcloud)

    # def config_video(self, ):
    # Nothing to configure for video recorder

    # TODO: implement config of BAG to either record depth as frame or pointcloud
    # def config_bag(self, pointcloud: bool):
    #     if self.type != RecordType.BAG:
    #         print(f"Recorder type is {self.type}, not BAG! Config attempt ignored.")
    #     self.recorder.set_pointcloud(pointcloud)


    def _createFolder(self, path: Path, mxid: str) -> Path:
        """
        Creates recording folder
        """
        i = 0
        while True:
            i += 1
            recordings_path = path / f"{i}-{str(mxid)}"
            if not recordings_path.is_dir():
                recordings_path.mkdir(parents=True, exist_ok=False)
                return recordings_path

    def close(self):
        if self.frame_q:
            self.frame_q.put(None)  # Close recorder and stop the thread



import json
from pathlib import Path
import depthai as dai
import cv2
import numpy as np

from .preview_manager import PreviewManager, SyncedPreviewManager
from ..previews import Previews
from ..utils import loadModule, toTensorResult, frameNorm, toPlanar


class NNetManager:
    """
    Manager class handling all NN-related functionalities. It's capable of creating appropriate nodes and connections,
    decoding neural network output automatically or by using external handler file.
    """

    def __init__(self, inputSize, nnFamily=None, labels=[], confidence=0.5, sync=False):
        """
        Args:
            inputSize (tuple): Desired NN input size, should match the input size defined in the network itself (width, height)
            nnFamily (str, Optional): type of NeuralNetwork to be processed. Supported: :code:`"YOLO"` and :code:`mobilenet`
            labels (list, Optional): Allows to display class label instead of ID when drawing nn detections.
            confidence (float, Optional): Specify detection nn's confidence threshold
            sync (bool, Optional): Store NN results for preview syncing (to be used with SyncedPreviewManager
        """
        self.inputSize = inputSize
        self._nnFamily = nnFamily
        if nnFamily in ("YOLO", "mobilenet"):
            self._outputFormat = "detection"
        self._labels = labels
        self._confidence = confidence
        self._sync = sync

    #: list: List of available neural network inputs
    sourceChoices = ("color", "left", "right", "rectifiedLeft", "rectifiedRight", "host")
    #: str: Selected neural network input
    source = None
    #: tuple: NN input size (width, height)
    inputSize = None
    #: depthai.OpenVINO.Version: OpenVINO version, available only if parsed from config file (see :func:`readConfig`)
    openvinoVersion = None
    #: depthai.DataInputQueue: DepthAI input queue object that allows to send images from host to device (used only with :code:`host` source)
    inputQueue = None
    #: depthai.DataOutputQueue: DepthAI output queue object that allows to receive NN results from the device.
    outputQueue = None
    #: dict: nn data buffer, disabled by default. Stores parsed nn data with packet sequence number as dict key
    buffer = {}


    _bboxColors = np.random.random(size=(256, 3)) * 256  # Random Colors for bounding boxes
    _countLabel = None
    _textBgColor = (0, 0, 0)
    _textColor = (255, 255, 255)
    _lineType = cv2.LINE_AA
    _textType = cv2.FONT_HERSHEY_SIMPLEX
    _outputFormat = "raw"
    _metadata = None
    _fullFov = True
    _config = None
    _nnFamily = None
    _handler = None

    def readConfig(self, path):
        """
        Parses the model config file and adjusts NNetManager values accordingly. It's advised to create a config file
        for every new network, as it allows to use dedicated NN nodes (for `MobilenetSSD <https://github.com/luxonis/depthai/blob/main/resources/nn/mobilenet-ssd/mobilenet-ssd.json>`__ and `YOLO <https://github.com/luxonis/depthai/blob/main/resources/nn/tiny-yolo-v3/tiny-yolo-v3.json>`__)
        or use `custom handler <https://github.com/luxonis/depthai/blob/main/resources/nn/openpose2/openpose2.json>`__ to process and display custom network results

        Args:
            path (pathlib.Path): Path to model config file (.json)

        Raises:
            ValueError: If path to config file does not exist
            RuntimeError: If custom handler does not contain :code:`draw` or :code:`show` methods
        """
        configPath = Path(path)
        if not configPath.exists():
            raise ValueError("Path {} does not exist!".format(path))

        with configPath.open() as f:
            self._config = json.load(f)
            if "openvino_version" in self._config:
                self.openvinoVersion =getattr(dai.OpenVINO.Version, 'VERSION_' + self._config.get("openvino_version"))
            nnConfig = self._config.get("nn_config", {})
            self._labels = self._config.get("mappings", {}).get("labels", None)
            self._nnFamily = nnConfig.get("NN_family", None)
            self._outputFormat = nnConfig.get("output_format", "raw")
            self._metadata = nnConfig.get("NN_specific_metadata", {})
            if "input_size" in nnConfig:
                self.inputSize = tuple(map(int, nnConfig.get("input_size").split('x')))

            self._confidence = self._metadata.get("confidence_threshold", nnConfig.get("confidence_threshold", None))
            if 'handler' in self._config:
                self._handler = loadModule(configPath.parent / self._config["handler"])

                if not callable(getattr(self._handler, "draw", None)) or not callable(getattr(self._handler, "decode", None)):
                    raise RuntimeError("Custom model handler does not contain 'draw' or 'decode' methods!")


    def _normFrame(self, frame):
        if not self._fullFov:
            scaleF = frame.shape[0] / self.inputSize[1]
            return np.zeros((int(self.inputSize[1] * scaleF), int(self.inputSize[0] * scaleF)))
        else:
            return frame

    def _cropOffsetX(self, frame):
        if not self._fullFov:
            croppedW = (frame.shape[0] / self.inputSize[1]) * self.inputSize[0]
            return int((frame.shape[1] - croppedW) // 2)
        else:
            return 0

    def createNN(self, pipeline, nodes, blobPath, source="color", useDepth=False, minDepth=100, maxDepth=10000, sbbScaleFactor=0.3, fullFov=True, useImageManip=True):
        """
        Creates nodes and connections in provided pipeline that will allow to run NN model and consume it's results.

        Args:
            pipeline (depthai.Pipeline): Pipeline instance
            nodes (types.SimpleNamespace): Object cointaining all of the nodes added to the pipeline. Available in :attr:`depthai_sdk.managers.PipelineManager.nodes`
            blobPath (pathlib.Path): Path to MyriadX blob. Might be useful to use together with
                :func:`depthai_sdk.managers.BlobManager.getBlob()` for dynamic blob compilation
            source (str, Optional): Neural network input source, one of :attr:`sourceChoices`
            useDepth (bool, Optional): If set to True, produced detections will have spatial coordinates included
            minDepth (int, Optional): Minimum depth distance in centimeters
            maxDepth (int, Optional): Maximum depth distance in centimeters
            sbbScaleFactor (float, Optional): Scale of the bounding box that will be used to calculate spatial coordinates for
                detection. If set to 0.3, it will scale down center-wise the bounding box to 0.3 of it's original size
                and use it to calculate spatial location of the object
            fullFov (bool, Optional): If set to False, manager will include crop offset when scaling the detections.
                Usually should be set to True (if you don't perform aspect ratio crop or when `keepAspectRatio` flag
                on camera/manip node is set to False
            useImageManip (bool, Optional): If set to False, manager will not create an image manip node for input image
                scaling - which may result in an input image being not adjusted for the NeuralNetwork node. Can be useful
                when we want to limit the amount of nodes running simultaneously on device

        Returns:
            depthai.node.NeuralNetwork: Configured NN node that was added to the pipeline

        Raises:
            RuntimeError: If source is not a valid choice or when input size has not been set.
        """
        if source not in self.sourceChoices:
            raise RuntimeError(f"Source {source} is invalid, available {self.sourceChoices}")
        if self.inputSize is None:
            raise RuntimeError("Unable to determine the nn input size. Please use --cnnInputSize flag to specify it in WxH format: -nnSize <width>x<height>")

        self.source = source
        self._fullFov = fullFov
        if self._nnFamily == "mobilenet":
            nodes.nn = pipeline.createMobileNetSpatialDetectionNetwork() if useDepth else pipeline.createMobileNetDetectionNetwork()
            nodes.nn.setConfidenceThreshold(self._confidence)
        elif self._nnFamily == "YOLO":
            nodes.nn = pipeline.createYoloSpatialDetectionNetwork() if useDepth else pipeline.createYoloDetectionNetwork()
            nodes.nn.setConfidenceThreshold(self._confidence)
            nodes.nn.setNumClasses(self._metadata["classes"])
            nodes.nn.setCoordinateSize(self._metadata["coordinates"])
            nodes.nn.setAnchors(self._metadata["anchors"])
            nodes.nn.setAnchorMasks(self._metadata["anchor_masks"])
            nodes.nn.setIouThreshold(self._metadata["iou_threshold"])
        else:
            # TODO use createSpatialLocationCalculator
            nodes.nn = pipeline.createNeuralNetwork()

        nodes.nn.setBlobPath(str(blobPath))
        nodes.nn.setNumInferenceThreads(2)
        nodes.nn.input.setBlocking(False)
        nodes.nn.input.setQueueSize(2)

        nodes.xoutNn = pipeline.createXLinkOut()
        nodes.xoutNn.setStreamName("nnOut")
        nodes.nn.out.link(nodes.xoutNn.input)

        if self.source == "host":
            nodes.xinNn = pipeline.createXLinkIn()
            nodes.xinNn.setMaxDataSize(self.inputSize[0] * self.inputSize[1] * 3)
            nodes.xinNn.setStreamName("nnIn")
            nodes.xinNn.out.link(nodes.nn.input)
        else:
            if useImageManip:
                nodes.manipNn = pipeline.createImageManip()
                nodes.manipNn.initialConfig.setResize(*self.inputSize)
                # The NN model expects BGR input. By default ImageManip output type would be same as input (gray in this case)
                nodes.manipNn.initialConfig.setFrameType(dai.RawImgFrame.Type.BGR888p)
                # NN inputs
                nodes.manipNn.out.link(nodes.nn.input)
                nodes.manipNn.setKeepAspectRatio(not self._fullFov)
                nodes.manipNn.setMaxOutputFrameSize(self.inputSize[0] * self.inputSize[1] * 3)

                link_input = nodes.manipNn.inputImage
            else:
                link_input = nodes.nn.input

            if self.source == "color":
                nodes.camRgb.preview.link(link_input)
            if self.source == "left":
                nodes.monoLeft.out.link(link_input)
            elif self.source == "right":
                nodes.monoRight.out.link(link_input)
            elif self.source == "rectifiedLeft":
                nodes.stereo.rectifiedLeft.link(link_input)
            elif self.source == "rectifiedRight":
                nodes.stereo.rectifiedRight.link(link_input)

        if self._nnFamily in ("YOLO", "mobilenet") and useDepth:
            nodes.stereo.depth.link(nodes.nn.inputDepth)
            nodes.nn.setDepthLowerThreshold(minDepth)
            nodes.nn.setDepthUpperThreshold(maxDepth)
            nodes.nn.setBoundingBoxScaleFactor(sbbScaleFactor)

        return nodes.nn

    def getLabelText(self, label):
        """
        Retrieves text assigned to specific label

        Args:
            label (int): Integer representing detection label, usually returned from NN node

        Returns:
            str: Label text assigned to specific label id or label id

        Raises:
            RuntimeError: If source is not a valid choice or when input size has not been set.
        """
        if self._labels is None:
            return str(label)
        elif int(label) < len(self._labels):
            return self._labels[int(label)]
        else:
            print(f"Label of ouf bounds (label index: {label}, available labels: {len(self._labels)}")
            return str(label)

    def parse(self, blocking=False):
        if self.outputQueue is None:
            return None, None

        if blocking:
            inNn = self.outputQueue.get()
        else:
            inNn = self.outputQueue.tryGet()
        if inNn is not None:
            data = self.decode(inNn)
            if self._sync:
                self.buffer[inNn.getSequenceNum()] = data
            return data, inNn
        else:
            return None, None

    def decode(self, inNn):
        """
        Decodes NN output. Performs generic handling for supported detection networks or calls custom handler methods

        Args:
            inNn (depthai.NNData): Integer representing detection label, usually returned from NN node

        Returns:
            Decoded NN data

        Raises:
            RuntimeError: if outputFormat specified in model config file is not recognized
        """
        if self._outputFormat == "detection":
            return inNn.detections
        elif self._outputFormat == "raw":
            if self._handler is not None:
                return self._handler.decode(self, inNn)
            else:
                try:
                    data = toTensorResult(inNn)
                    print("Received NN packet: ", ", ".join([f"{key}: {value.shape}" for key, value in data.items()]))
                except Exception as ex:
                    print("Received NN packet: <Preview unabailable: {}>".format(ex))
        else:
            raise RuntimeError("Unknown output format: {}".format(self._outputFormat))

    def _drawCount(self, source, decodedData):
        def drawCnt(frame, cnt):
            cv2.putText(frame, f"{self._countLabel}: {cnt}", (5, 46), self._textType, 0.5, self._textBgColor, 4, self._lineType)
            cv2.putText(frame, f"{self._countLabel}: {cnt}", (5, 46), self._textType, 0.5, self._textColor, 1, self._lineType)

        # Count the number of detected objects
        cntList = list(filter(lambda x: self.getLabelText(x.label) == self._countLabel, decodedData))
        if isinstance(source, PreviewManager):
            for frame in source.frames.values():
                drawCnt(frame, len(cntList))
        else:
            drawCnt(source, len(cntList))

    def draw(self, source, decodedData):
        """
        Draws NN results onto the frames. It's responsible to correctly map the results onto each frame requested,
        including applying crop offset or preparing a correct normalization frame, then draws them with all information
        provided (confidence, label, spatial location, label count).

        Also, it's able to call custom nn handler method :code:`draw` to hand over drawing the results

        Args:
            source (depthai_sdk.managers.PreviewManager | numpy.ndarray): Draw target.
                If supplied with a regular frame, it will draw the count on that frame

                If supplied with :class:`depthai_sdk.managers.PreviewManager` instance, it will print the count label
                on all of the frames that it stores

            decodedData: Detections from neural network node, usually returned from :func:`decode` method
        """
        if self._outputFormat == "detection":
            def drawDetection(frame, detection):
                bbox = frameNorm(self._normFrame(frame), [detection.xmin, detection.ymin, detection.xmax, detection.ymax])
                if self.source == Previews.color.name and not self._fullFov:
                    bbox[::2] += self._cropOffsetX(frame)
                cv2.rectangle(frame, (bbox[0], bbox[1]), (bbox[2], bbox[3]), self._bboxColors[detection.label], 2)
                cv2.rectangle(frame, (bbox[0], (bbox[1] - 28)), ((bbox[0] + 110), bbox[1]), self._bboxColors[detection.label], cv2.FILLED)
                cv2.putText(frame, self.getLabelText(detection.label), (bbox[0] + 5, bbox[1] - 10),
                            self._textType, 0.5, (0, 0, 0), 1, self._lineType)
                cv2.putText(frame, f"{int(detection.confidence * 100)}%", (bbox[0] + 62, bbox[1] - 10),
                            self._textType, 0.5, (0, 0, 0), 1, self._lineType)

                if hasattr(detection, 'spatialCoordinates'):  # Display spatial coordinates as well
                    xMeters = detection.spatialCoordinates.x / 1000
                    yMeters = detection.spatialCoordinates.y / 1000
                    zMeters = detection.spatialCoordinates.z / 1000
                    cv2.putText(frame, "X: {:.2f} m".format(xMeters), (bbox[0] + 10, bbox[1] + 60),
                                self._textType, 0.5, self._textBgColor, 4, self._lineType)
                    cv2.putText(frame, "X: {:.2f} m".format(xMeters), (bbox[0] + 10, bbox[1] + 60),
                                self._textType, 0.5, self._textColor, 1, self._lineType)
                    cv2.putText(frame, "Y: {:.2f} m".format(yMeters), (bbox[0] + 10, bbox[1] + 75),
                                self._textType, 0.5, self._textBgColor, 4, self._lineType)
                    cv2.putText(frame, "Y: {:.2f} m".format(yMeters), (bbox[0] + 10, bbox[1] + 75),
                                self._textType, 0.5, self._textColor, 1, self._lineType)
                    cv2.putText(frame, "Z: {:.2f} m".format(zMeters), (bbox[0] + 10, bbox[1] + 90),
                                self._textType, 0.5, self._textBgColor, 4, self._lineType)
                    cv2.putText(frame, "Z: {:.2f} m".format(zMeters), (bbox[0] + 10, bbox[1] + 90),
                                self._textType, 0.5, self._textColor, 1, self._lineType)
            if isinstance(source, SyncedPreviewManager):
                if len(self.buffer) > 0 and source.nnSyncSeq is not None:
                    data = self.buffer.get(source.nnSyncSeq, self.buffer[max(self.buffer.keys())])
                    for old_key in list(filter(lambda key: key < source.nnSyncSeq, self.buffer.keys())):
                        del self.buffer[old_key]
                    for detection in data:
                        for name, frame in source.frames.items():
                            drawDetection(frame, detection)
            else:
                for detection in decodedData:
                    if isinstance(source, PreviewManager):
                        for name, frame in source.frames.items():
                            drawDetection(frame, detection)
                    else:
                        drawDetection(source, detection)

            if self._countLabel is not None:
                self._drawCount(source, decodedData)

        elif self._outputFormat == "raw" and self._handler is not None:
            if isinstance(source, PreviewManager):
                frames = list(source.frames.items())
            else:
                frames = [("host", source)]
            self._handler.draw(self, decodedData, frames)

    def createQueues(self, device):
        """
        Creates output queue for NeuralNetwork node and, if using :code:`host` as a :attr:`source`, it will also create
        input queue.

        Args:
            device (depthai.Device): Running device instance
        """
        if self.source == "host":
            self.inputQueue = device.getInputQueue("nnIn", maxSize=1, blocking=False)
        self.outputQueue = device.getOutputQueue("nnOut", maxSize=1, blocking=False)

    def closeQueues(self):
        """
        Closes output queues created by :func:`createQueues`
        """
        if self.source == "host" and self.inputQueue is not None:
            self.inputQueue.close()
        if self.outputQueue is not None:
            self.outputQueue.close()

    def sendInputFrame(self, frame, seqNum=None):
        """
        Sends a frame into :attr:`inputQueue` object. Handles scaling down the frame, creating a proper :obj:`depthai.ImgFrame`
        and sending it to the queue. Be sure to use :code:`host` as a :attr:`source` and call :func:`createQueues` prior
        input queue.

        Args:
            frame (numpy.ndarray): Frame to be sent to the device
            seqNum (int, Optional): Sequence number set on ImgFrame. Useful in synchronization scenarios

        Returns:
            numpy.ndarray: scaled frame that was sent to the NN (same width/height as NN input)

        Raises:
            RuntimeError: if :attr:`inputQueue` is :code:`None` (unable to send the image)
        """
        if self.inputQueue is None:
            raise RuntimeError("Unable to send image, no input queue is present! Call `createQueues(device)` first!")

        scaledFrame = cv2.resize(frame, self.inputSize)
        frameNn = dai.ImgFrame()
        if seqNum is not None:
            frameNn.setSequenceNum(seqNum)
        frameNn.setType(dai.ImgFrame.Type.BGR888p)
        frameNn.setWidth(self.inputSize[0])
        frameNn.setHeight(self.inputSize[1])
        frameNn.setData(toPlanar(scaledFrame))
        self.inputQueue.send(frameNn)

        return scaledFrame

    def countLabel(self, label):
        """
        Enables object count for specific label. Label count will be printed once :func:`draw` method is called

        Args:
            label (str | int): Label to be counted. If model is using mappings in model config file, supply here a :obj:`str` label
                to be tracked. If no mapping is present, specify the label as :obj:`int` (NN-default)
        """

        self._countLabel = label


import time

class FPS:
    def __init__(self):
        self.timestamp = time.time() + 1
        self.start = time.time()
        self.frame_cnt = 0

    def next_iter(self):
        self.timestamp = time.time()
        self.frame_cnt += 1

    def fps(self) -> float:
        diff = self.timestamp - self.start
        return self.frame_cnt / diff if diff != 0 else 0.0



from typing import Union, Tuple, Optional, Dict, Any, Type

import depthai as dai


def rgb_resolution(resolution: Union[
    None, str, dai.ColorCameraProperties.SensorResolution]) -> dai.ColorCameraProperties.SensorResolution:
    """
    Parses Color camera resolution based on the string
    """
    if isinstance(resolution, dai.ColorCameraProperties.SensorResolution) or resolution is None:
        return resolution

    resolution = str(resolution).upper()
    if resolution == '3120' or resolution == '13MP':
        return dai.ColorCameraProperties.SensorResolution.THE_13_MP
    elif resolution == '3040' or resolution == '12MP':
        return dai.ColorCameraProperties.SensorResolution.THE_12_MP
    elif resolution == '2160' or resolution == '4K':
        return dai.ColorCameraProperties.SensorResolution.THE_4_K
    # elif resolution == '1920' or resolution == '5MP':
    #     return dai.ColorCameraProperties.SensorResolution.THE_5_MP
    elif resolution == '800' or resolution == '800P':
        return dai.ColorCameraProperties.SensorResolution.THE_800_P
    elif resolution == '720' or resolution == '720P':
        return dai.ColorCameraProperties.SensorResolution.THE_720_P
    else:  # Default
        return dai.ColorCameraProperties.SensorResolution.THE_1080_P


def mono_resolution(resolution: Union[
    None, str, dai.MonoCameraProperties.SensorResolution]) -> dai.MonoCameraProperties.SensorResolution:
    """
    Parses Mono camera resolution based on the string
    """
    if isinstance(resolution, dai.MonoCameraProperties.SensorResolution) or resolution is None:
        return resolution

    resolution = str(resolution).upper()
    if resolution == '800' or resolution == '800P' or resolution == '1MP':
        return dai.MonoCameraProperties.SensorResolution.THE_800_P
    elif resolution == '720' or resolution == '720P':
        return dai.MonoCameraProperties.SensorResolution.THE_720_P
    elif resolution == '480' or resolution == '480P':
        return dai.MonoCameraProperties.SensorResolution.THE_480_P
    else:  # Default
        return dai.MonoCameraProperties.SensorResolution.THE_400_P


def parse_resolution(
        camera: Union[dai.node.ColorCamera, dai.node.MonoCamera, Type],
        resolution: Union[str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution]
):
    if isinstance(camera, dai.node.ColorCamera) or camera == dai.node.ColorCamera:
        return rgb_resolution(resolution)
    elif isinstance(camera, type(dai.node.MonoCamera)) or camera == dai.node.ColorCamera:
        return mono_resolution(resolution)
    else:
        raise ValueError("camera must be either MonoCamera or ColorCamera!")


def parse_bool(value: str) -> bool:
    if value.upper() in ['1', 'TRUE', 'ON', 'YES']:
        return True
    elif value.upper() in ['0', 'FALSE', 'OFF', 'NO']:
        return False
    else:
        raise ValueError(f"Couldn't parse '{value}' to bool!")


def parse_usb_speed(speed: Union[None, str, dai.UsbSpeed]) -> Optional[dai.UsbSpeed]:
    if speed is None:
        return None
    elif isinstance(speed, dai.UsbSpeed):
        return speed
    elif isinstance(speed, str):
        if speed.upper() in ['HIGH', '2', 'USB2']:
            return dai.UsbSpeed.HIGH
        elif speed.upper() in ['SUPER', '3', 'USB3']:
            return dai.UsbSpeed.SUPER
    raise ValueError(f"Could not parse USB speed '{speed}' to dai.UsbSpeed!")


def parse_median_filter(filter: Union[int, dai.MedianFilter]) -> dai.MedianFilter:
    if isinstance(filter, dai.MedianFilter):
        return filter

    if filter == 3:
        return dai.MedianFilter.KERNEL_3x3
    elif filter == 5:
        return dai.MedianFilter.KERNEL_5x5
    elif filter == 7:
        return dai.MedianFilter.KERNEL_7x7
    else:
        return dai.MedianFilter.MEDIAN_OFF


def parse_open_vino_version(version: Union[None, str, dai.OpenVINO.Version]) -> Optional[dai.OpenVINO.Version]:
    if version is None:
        return None
    if isinstance(version, str):
        vals = None
        if '.' in version:
            vals = version.split('.')
        elif '_' in version:
            vals = version.split('_')
        if vals is None:
            return None
        version = getattr(dai.OpenVINO.Version, f"VERSION_{vals[0]}_{vals[1]}")
    return version


def parse_size(size: Union[str, Tuple[int, int]]) -> Tuple[int, int]:
    if isinstance(size, Tuple):
        return size
    elif isinstance(size, str):
        vals = size.split('x')
        if len(vals) != 2: raise ValueError("Size must be in format '[width]x[height]'!")
        return (int(vals[0]), int(vals[1]))
    else:
        raise ValueError("Size type not supported!")


def parse_color_cam_control(options: Dict[str, Any], cam: dai.node.ColorCamera):
    from .camera_helper import setCameraControl
    setCameraControl(cam.initialControl,
                     options.get('manualFocus', None),
                     options.get('afMode', None),
                     options.get('awbMode', None),
                     options.get('sceneMode', None),
                     options.get('antiBandingMode', None),
                     options.get('effectMode', None),
                     options.get('sharpness', None),
                     options.get('lumaDenoise', None),
                     options.get('chromaDenoise', None),
                     )


def parse_encode(encode=Union[str, bool, dai.VideoEncoderProperties.Profile]
                 ) -> dai.VideoEncoderProperties.Profile:
    if isinstance(encode, dai.VideoEncoderProperties.Profile):
        return encode
    elif isinstance(encode, bool) and encode:
        return dai.VideoEncoderProperties.Profile.MJPEG  # MJPEG by default
    elif isinstance(encode, str):
        encode = encode.upper()
        if encode in ['MJPEG', 'JPEG', 'JPG', '.MJPEG', '.JPEG', '.JPG']:
            return dai.VideoEncoderProperties.Profile.MJPEG
        elif encode in ['H265', '.H265', 'H.265', 'HEVC', '.HEVC']:
            return dai.VideoEncoderProperties.Profile.H265_MAIN
        elif encode in ['H264', '.H264', 'H.264', 'MPEG-4', 'MPEG', 'AVC']:
            return dai.VideoEncoderProperties.Profile.H264_MAIN
    raise ValueError(f"Parsing user-defined encode value '{encode}' failed!")


def parse_cam_socket(socket=Union[str, dai.CameraBoardSocket]) -> dai.CameraBoardSocket:
    if isinstance(socket, dai.CameraBoardSocket):
        return socket
    elif isinstance(socket, str):
        socket = socket.upper()
        if socket in ['RGB', 'COLOR']:
            return dai.CameraBoardSocket.RGB
        elif socket == 'RIGHT':
            return dai.CameraBoardSocket.RIGHT
        elif socket == 'LEFT':
            return dai.CameraBoardSocket.LEFT
    raise ValueError(f"Parsing user-defined camera board socket value '{socket}' failed!")



#!/usr/bin/python
import math

from Qt import QtCore, QtGui, QtWidgets

from ..constants import (
    PipeEnum, PipeLayoutEnum, PortTypeEnum, Z_VAL_PIPE,
    Z_VAL_NODE_WIDGET,
    ITEM_CACHE_MODE,
    NODE_LAYOUT_VERTICAL, NODE_LAYOUT_HORIZONTAL,
    NODE_LAYOUT_DIRECTION
)
from ..qgraphics.port import PortItem

PIPE_STYLES = {
    PipeEnum.DRAW_TYPE_DEFAULT.value: QtCore.Qt.SolidLine,
    PipeEnum.DRAW_TYPE_DASHED.value: QtCore.Qt.DashLine,
    PipeEnum.DRAW_TYPE_DOTTED.value: QtCore.Qt.DotLine
}


class PipeItem(QtWidgets.QGraphicsPathItem):
    """
    Base Pipe item used for drawing node connections.
    """

    def __init__(self, input_port=None, output_port=None):
        super(PipeItem, self).__init__()
        self.setZValue(Z_VAL_PIPE)
        self.setAcceptHoverEvents(True)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable)
        self._color = PipeEnum.COLOR.value
        self._style = PipeEnum.DRAW_TYPE_DEFAULT.value
        self._active = False
        self._highlight = False
        self._input_port = input_port
        self._output_port = output_port
        size = 6.0
        self._arrow = QtGui.QPolygonF()
        self._arrow.append(QtCore.QPointF(-size, size))
        self._arrow.append(QtCore.QPointF(0.0, -size * 1.5))
        self._arrow.append(QtCore.QPointF(size, size))
        self.setCacheMode(ITEM_CACHE_MODE)

    def __repr__(self):
        in_name = self._input_port.name if self._input_port else ''
        out_name = self._output_port.name if self._output_port else ''
        return '{}.Pipe(\'{}\', \'{}\')'.format(
            self.__module__, in_name, out_name)

    def hoverEnterEvent(self, event):
        self.activate()

    def hoverLeaveEvent(self, event):
        self.reset()
        if self.input_port and self.output_port:
            if self.input_port.node.selected:
                self.highlight()
            elif self.output_port.node.selected:
                self.highlight()
        if self.isSelected():
            self.highlight()

    def paint(self, painter, option, widget):
        """
        Draws the connection line between nodes.

        Args:
            painter (QtGui.QPainter): painter used for drawing the item.
            option (QtGui.QStyleOptionGraphicsItem):
                used to describe the parameters needed to draw.
            widget (QtWidgets.QWidget): not used.
        """
        color = QtGui.QColor(*self._color)
        pen_style = PIPE_STYLES.get(self.style)
        pen_width = PipeEnum.WIDTH.value
        if self._active:
            # GX
            pen_width = PipeEnum.ACTIVE_WIDTH.value
            color = QtGui.QColor(*PipeEnum.ACTIVE_COLOR.value)
            if pen_style == QtCore.Qt.DashDotDotLine:
                pen_width += 1
            else:
                pen_width += 0.35
        elif self._highlight:
            color = QtGui.QColor(*PipeEnum.HIGHLIGHT_COLOR.value)
            pen_style = PIPE_STYLES.get(PipeEnum.DRAW_TYPE_DEFAULT.value)
            # GX
            pen_width = PipeEnum.HIGHLIGHT_WIDTH.value


        if self.disabled():
            if not self._active:
                color = QtGui.QColor(*PipeEnum.DISABLED_COLOR.value)
            pen_width += 0.2
            pen_style = PIPE_STYLES.get(PipeEnum.DRAW_TYPE_DOTTED.value)

        pen = QtGui.QPen(color, pen_width, pen_style)
        pen.setCapStyle(QtCore.Qt.RoundCap)
        pen.setJoinStyle(QtCore.Qt.MiterJoin)

        painter.save()
        painter.setPen(pen)
        painter.setRenderHint(painter.Antialiasing, True)
        painter.drawPath(self.path())

        # draw arrow
        if self.input_port and self.output_port:
            cen_x = self.path().pointAtPercent(0.5).x()
            cen_y = self.path().pointAtPercent(0.5).y()
            loc_pt = self.path().pointAtPercent(0.49)
            tgt_pt = self.path().pointAtPercent(0.51)

            dist = math.hypot(tgt_pt.x() - cen_x, tgt_pt.y() - cen_y)
            if dist < 0.5:
                painter.restore()
                return

            color.setAlpha(255)
            if self._highlight:
                painter.setBrush(QtGui.QBrush(color.lighter(150)))
            elif self._active or self.disabled():
                painter.setBrush(QtGui.QBrush(color.darker(200)))
            else:
                painter.setBrush(QtGui.QBrush(color.darker(130)))

            pen_width = 0.6
            if dist < 1.0:
                pen_width *= (1.0 + dist)

            pen = QtGui.QPen(color, pen_width)
            pen.setCapStyle(QtCore.Qt.RoundCap)
            pen.setJoinStyle(QtCore.Qt.MiterJoin)
            painter.setPen(pen)

            transform = QtGui.QTransform()
            transform.translate(cen_x, cen_y)
            radians = math.atan2(tgt_pt.y() - loc_pt.y(),
                                 tgt_pt.x() - loc_pt.x())
            degrees = math.degrees(radians) - 90
            transform.rotate(degrees)
            if dist < 1.0:
                transform.scale(dist, dist)
            painter.drawPolygon(transform.map(self._arrow))

        # QPaintDevice: Cannot destroy paint device that is being painted.
        painter.restore()

    def __draw_path_vertical(self, start_port, pos1, pos2, path):
        """
        Draws the vertical path between ports.

        Args:
            start_port (PortItem): port used to draw the starting point.
            pos1 (QPointF): start port position.
            pos2 (QPointF): end port position.
            path (QPainterPath): path to draw.
        """
        if self.viewer_pipe_layout() == PipeLayoutEnum.CURVED.value:
            ctr_offset_y1, ctr_offset_y2 = pos1.y(), pos2.y()
            tangent = abs(ctr_offset_y1 - ctr_offset_y2)

            max_height = start_port.node.boundingRect().height()
            tangent = min(tangent, max_height)
            if start_port.port_type == PortTypeEnum.IN.value:
                ctr_offset_y1 -= tangent
                ctr_offset_y2 += tangent
            else:
                ctr_offset_y1 += tangent
                ctr_offset_y2 -= tangent

            ctr_point1 = QtCore.QPointF(pos1.x(), ctr_offset_y1)
            ctr_point2 = QtCore.QPointF(pos2.x(), ctr_offset_y2)
            path.cubicTo(ctr_point1, ctr_point2, pos2)
            self.setPath(path)
        elif self.viewer_pipe_layout() == PipeLayoutEnum.ANGLE.value:
            ctr_offset_y1, ctr_offset_y2 = pos1.y(), pos2.y()
            distance = abs(ctr_offset_y1 - ctr_offset_y2)/2
            if start_port.port_type == PortTypeEnum.IN.value:
                ctr_offset_y1 -= distance
                ctr_offset_y2 += distance
            else:
                ctr_offset_y1 += distance
                ctr_offset_y2 -= distance

            ctr_point1 = QtCore.QPointF(pos1.x(), ctr_offset_y1)
            ctr_point2 = QtCore.QPointF(pos2.x(), ctr_offset_y2)
            path.lineTo(ctr_point1)
            path.lineTo(ctr_point2)
            path.lineTo(pos2)
            self.setPath(path)

    def __draw_path_horizontal(self, start_port, pos1, pos2, path):
        """
        Draws the horizontal path between ports.

        Args:
            start_port (PortItem): port used to draw the starting point.
            pos1 (QPointF): start port position.
            pos2 (QPointF): end port position.
            path (QPainterPath): path to draw.
        """
        if self.viewer_pipe_layout() == PipeLayoutEnum.CURVED.value:
            ctr_offset_x1, ctr_offset_x2 = pos1.x(), pos2.x()
            tangent = abs(ctr_offset_x1 - ctr_offset_x2)

            max_width = start_port.node.boundingRect().width()
            tangent = min(tangent, max_width)
            if start_port.port_type == PortTypeEnum.IN.value:
                ctr_offset_x1 -= tangent
                ctr_offset_x2 += tangent
            else:
                ctr_offset_x1 += tangent
                ctr_offset_x2 -= tangent

            ctr_point1 = QtCore.QPointF(ctr_offset_x1, pos1.y())
            ctr_point2 = QtCore.QPointF(ctr_offset_x2, pos2.y())
            path.cubicTo(ctr_point1, ctr_point2, pos2)
            self.setPath(path)
        elif self.viewer_pipe_layout() == PipeLayoutEnum.ANGLE.value:
            ctr_offset_x1, ctr_offset_x2 = pos1.x(), pos2.x()
            distance = abs(ctr_offset_x1 - ctr_offset_x2) / 2
            if start_port.port_type == PortTypeEnum.IN.value:
                ctr_offset_x1 -= distance
                ctr_offset_x2 += distance
            else:
                ctr_offset_x1 += distance
                ctr_offset_x2 -= distance

            ctr_point1 = QtCore.QPointF(ctr_offset_x1, pos1.y())
            ctr_point2 = QtCore.QPointF(ctr_offset_x2, pos2.y())
            path.lineTo(ctr_point1)
            path.lineTo(ctr_point2)
            path.lineTo(pos2)
            self.setPath(path)

    def draw_path(self, start_port, end_port=None, cursor_pos=None):
        """
        Draws the path between ports.

        Args:
            start_port (PortItem): port used to draw the starting point.
            end_port (PortItem): port used to draw the end point.
            cursor_pos (QtCore.QPointF): cursor position if specified this
                will be the draw end point.
        """
        if not start_port:
            return
        pos1 = start_port.scenePos()
        pos1.setX(pos1.x() + (start_port.boundingRect().width() / 2))
        pos1.setY(pos1.y() + (start_port.boundingRect().height() / 2))
        if cursor_pos:
            pos2 = cursor_pos
        elif end_port:
            pos2 = end_port.scenePos()
            pos2.setX(pos2.x() + (start_port.boundingRect().width() / 2))
            pos2.setY(pos2.y() + (start_port.boundingRect().height() / 2))
        else:
            return

        line = QtCore.QLineF(pos1, pos2)
        path = QtGui.QPainterPath()
        path.moveTo(line.x1(), line.y1())

        if self.viewer_pipe_layout() == PipeLayoutEnum.STRAIGHT.value:
            path.lineTo(pos2)
            self.setPath(path)
            return
        else:
            if NODE_LAYOUT_DIRECTION is NODE_LAYOUT_VERTICAL:
                self.__draw_path_vertical(start_port, pos1, pos2, path)
            elif NODE_LAYOUT_DIRECTION is NODE_LAYOUT_HORIZONTAL:
                self.__draw_path_horizontal(start_port, pos1, pos2, path)

    def reset_path(self):
        path = QtGui.QPainterPath(QtCore.QPointF(0.0, 0.0))
        self.setPath(path)

    @staticmethod
    def calc_distance(p1, p2):
        x = math.pow((p2.x() - p1.x()), 2)
        y = math.pow((p2.y() - p1.y()), 2)
        return math.sqrt(x + y)

    def port_from_pos(self, pos, reverse=False):
        inport_pos = self.input_port.scenePos()
        outport_pos = self.output_port.scenePos()
        input_dist = self.calc_distance(inport_pos, pos)
        output_dist = self.calc_distance(outport_pos, pos)
        if input_dist < output_dist:
            port = self.output_port if reverse else self.input_port
        else:
            port = self.input_port if reverse else self.output_port
        return port

    def viewer_pipe_layout(self):
        if self.scene():
            viewer = self.scene().viewer()
            return viewer.get_pipe_layout()

    def activate(self):
        self._active = True
        color = QtGui.QColor(*PipeEnum.ACTIVE_COLOR.value)
        pen = QtGui.QPen(
            color, 2.5, PIPE_STYLES.get(PipeEnum.DRAW_TYPE_DEFAULT.value)
        )
        self.setPen(pen)

    def active(self):
        return self._active

    def highlight(self):
        self._highlight = True
        color = QtGui.QColor(*PipeEnum.HIGHLIGHT_COLOR.value)
        pen = QtGui.QPen(
            color, 2, PIPE_STYLES.get(PipeEnum.DRAW_TYPE_DEFAULT.value)
        )
        self.setPen(pen)

    def highlighted(self):
        return self._highlight

    def reset(self):
        self._active = False
        self._highlight = False
        color = QtGui.QColor(*self.color)
        pen = QtGui.QPen(color, 2, PIPE_STYLES.get(self.style))
        self.setPen(pen)

    def set_connections(self, port1, port2):
        ports = {
            port1.port_type: port1,
            port2.port_type: port2
        }
        self.input_port = ports[PortTypeEnum.IN.value]
        self.output_port = ports[PortTypeEnum.OUT.value]
        ports[PortTypeEnum.IN.value].add_pipe(self)
        ports[PortTypeEnum.OUT.value].add_pipe(self)

    def disabled(self):
        if self.input_port and self.input_port.node.disabled:
            return True
        if self.output_port and self.output_port.node.disabled:
            return True
        return False

    def itemChange(self, change, value):
        if change == self.ItemSelectedChange and self.scene():
            self.reset()
            if value:
                self.highlight()
        return super(PipeItem, self).itemChange(change, value)

    @property
    def input_port(self):
        return self._input_port

    @input_port.setter
    def input_port(self, port):
        if isinstance(port, PortItem) or not port:
            self._input_port = port
        else:
            self._input_port = None

    @property
    def output_port(self):
        return self._output_port

    @output_port.setter
    def output_port(self, port):
        if isinstance(port, PortItem) or not port:
            self._output_port = port
        else:
            self._output_port = None

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        self._color = color

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, style):
        self._style = style

    def delete(self):
        if self.input_port and self.input_port.connected_pipes:
            self.input_port.remove_pipe(self)
        if self.output_port and self.output_port.connected_pipes:
            self.output_port.remove_pipe(self)
        if self.scene():
            self.scene().removeItem(self)


class LivePipeItem(PipeItem):

    def __init__(self):
        super(LivePipeItem, self).__init__()
        self.setZValue(Z_VAL_NODE_WIDGET + 1)
        self.shift_selected = False

    def paint(self, painter, option, widget):
        """
        Draws the connection line.

        Args:
            painter (QtGui.QPainter): painter used for drawing the item.
            option (QtGui.QStyleOptionGraphicsItem):
                used to describe the parameters needed to draw.
            widget (QtWidgets.QWidget): not used.
        """
        color = QtGui.QColor(*PipeEnum.ACTIVE_COLOR.value)
        pen_style = PIPE_STYLES.get(PipeEnum.DRAW_TYPE_DASHED.value)
        pen_width = PipeEnum.WIDTH.value + 0.35

        pen = QtGui.QPen(color, pen_width)
        pen.setStyle(pen_style)
        pen.setCapStyle(QtCore.Qt.RoundCap)

        painter.save()
        painter.setPen(pen)
        painter.setRenderHint(painter.Antialiasing, True)
        painter.drawPath(self.path())

        cen_x = self.path().pointAtPercent(0.5).x()
        cen_y = self.path().pointAtPercent(0.5).y()
        loc_pt = self.path().pointAtPercent(0.9)
        tgt_pt = self.path().pointAtPercent(1.0)
        start_pt = self.path().pointAtPercent(0.0)

        dist = math.hypot(tgt_pt.x() - cen_x, tgt_pt.y() - cen_y)
        if dist < 0.05:
            painter.restore()
            return

        # draw start circle
        size = 5.0
        rect = QtCore.QRectF(start_pt.x() - (size / 2),
                             start_pt.y() - (size / 2),
                             size, size)
        painter.setBrush(color)
        painter.drawEllipse(rect)

        # draw middle circle
        size = 10.0
        if dist < 50.0:
            size *= (dist / 50.0)
        rect = QtCore.QRectF(cen_x-(size/2), cen_y-(size/2), size, size)
        painter.setBrush(color)
        painter.setPen(QtGui.QPen(color.darker(130), pen_width))
        painter.drawEllipse(rect)

        # draw arrow
        color.setAlpha(255)
        painter.setBrush(color.darker(200))

        pen_width = 0.6
        if dist < 1.0:
            pen_width *= 1.0 + dist
        painter.setPen(QtGui.QPen(color, pen_width))

        transform = QtGui.QTransform()
        transform.translate(tgt_pt.x(), tgt_pt.y())

        radians = math.atan2(tgt_pt.y() - loc_pt.y(),
                             tgt_pt.x() - loc_pt.x())
        degrees = math.degrees(radians) + 90
        transform.rotate(degrees)

        scale = 1.0
        if dist < 20.0:
            scale = dist / 20.0
        transform.scale(scale, scale)
        painter.drawPolygon(transform.map(self._arrow))
        painter.restore()



from typing import Dict, Any, Union

import numpy as np

from .abstract_recorder import *


class VideoRecorder(Recorder):
    """
    Writes encoded streams raw (.mjpeg/.h264/.hevc) or directly to mp4 container.
    Writes unencoded streams to mp4 using cv2.VideoWriter
    """
    _closed = False
    _writers: Dict[str, Any]

    def __init__(self):
        self.path = None
        self._stream_type = dict()
        self._writer = dict()

    def __getitem__(self, item):
        return self._writer[item]

    # TODO device is not used
    def update(self, path: Path, device: dai.Device, xouts: List['XoutFrames']):
        """
        Update the recorder with new streams.
        Args:
            path: Path to save the output. Either a folder or a file.
            device: Device to get the streams from.
            xouts: List of output streams.
        """
        if path is None:
            return

        self.path = path
        if path.suffix == '' and path != Path('.'):  # If no extension, create a folder
            self.path.mkdir(parents=True, exist_ok=True)

        for xout in xouts:
            name = xout.name
            stream = OakStream(xout)
            fourcc = stream.fourcc()  # TODO add default fourcc? stream.fourcc() can be None.
            if stream.isRaw():
                from .video_writers.video_writer import VideoWriter
                self._writer[name] = VideoWriter(self.path, name, fourcc, xout.fps)
            else:
                try:
                    from .video_writers.av_writer import AvWriter
                    self._writer[name] = AvWriter(self.path, name, fourcc, xout.fps)
                except Exception as e:
                    # TODO here can be other errors, not only import error
                    print('Exception while creating AvWriter: ', e)
                    print('Falling back to FileWriter, saving uncontainerized encoded streams.')
                    from .video_writers.file_writer import FileWriter
                    self._writer[name] = FileWriter(self.path, name, fourcc)

    def write(self, name: str, frame: Union[np.ndarray, dai.ImgFrame]):
        self._writer[name].write(frame)

    def add_to_buffer(self, name: str, frame: Union[np.ndarray, dai.ImgFrame]):
        self._writer[name].add_to_buffer(frame)

    def close(self):
        if self._closed: return
        self._closed = True
        print("Video Recorder saved stream(s) to folder:", str(self.path))
        # Close opened files
        for name, writer in self._writer.items():
            writer.close()



#!/usr/bin/env python3

import signal
from typing import Dict
from depthai_sdk.components.node_graph_qt import NodeGraph, BaseNode, PropertiesBinWidget
from depthai_sdk.components.node_graph_qt.constants import ViewerEnum


class DepthaiNode(BaseNode):
    # unique node identifier.
    __identifier__ = 'dai'

    # initial default node name.
    NODE_NAME = 'Node'

    def __init__(self):
        super(DepthaiNode, self).__init__()

        # create QLineEdit text input widget.
        # self.add_text_input('my_input', 'Text Input', tab='widgets')


class PipelineGraph:

    def __init__(self, schema: Dict):

        from Qt import QtWidgets, QtCore

        node_color = {
            "ColorCamera": (241, 148, 138),
            "MonoCamera": (243, 243, 243),
            "ImageManip": (174, 214, 241),
            "VideoEncoder": (190, 190, 190),

            "NeuralNetwork": (171, 235, 198),
            "DetectionNetwork": (171, 235, 198),
            "MobileNetDetectionNetwork": (171, 235, 198),
            "MobileNetSpatialDetectionNetwork": (171, 235, 198),
            "YoloDetectionNetwork": (171, 235, 198),
            "YoloSpatialDetectionNetwork": (171, 235, 198),
            "SpatialDetectionNetwork": (171, 235, 198),

            "SPIIn": (242, 215, 213),
            "XLinkIn": (242, 215, 213),

            "SPIOut": (230, 176, 170),
            "XLinkOut": (230, 176, 170),

            "Script": (249, 231, 159),

            "StereoDepth": (215, 189, 226),
            "SpatialLocationCalculator": (215, 189, 226),

            "EdgeDetector": (248, 196, 113),
            "FeatureTracker": (248, 196, 113),
            "ObjectTracker": (248, 196, 113),
            "IMU": (248, 196, 113)
        }

        default_node_color = (190, 190, 190)  # For node types that does not appear in 'node_color'

        # handle SIGINT to make the app terminate on CTRL+C
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)

        app = QtWidgets.QApplication(["DepthAI Pipeline Graph"])

        # create node graph controller.
        graph = NodeGraph()
        graph.set_background_color(255, 255, 255)
        graph.set_grid_mode(ViewerEnum.GRID_DISPLAY_NONE.value)

        graph.register_node(DepthaiNode)

        # create a node properties bin widget.
        properties_bin = PropertiesBinWidget(node_graph=graph)
        properties_bin.setWindowFlags(QtCore.Qt.Tool)

        # show the node properties bin widget when a node is double clicked.
        def display_properties_bin(node):
            global properties_bin
            if not properties_bin.isVisible():
                properties_bin.show()

        # wire function to "node_double_clicked" signal.
        graph.node_double_clicked.connect(display_properties_bin)

        # show the node graph widget.
        graph_widget = graph.widget
        graph_widget.resize(1100, 800)

        dai_connections = schema['connections']
        dai_nodes = {}  # key = id, value = dict with keys 'type', 'blocking', 'queue_size' and 'name' (if args.use_variable_name)
        for n in schema['nodes']:
            dict_n = n[1]
            dai_nodes[dict_n['id']] = {'type': dict_n['name']}
            dai_nodes[dict_n['id']]['name'] = f"{dict_n['name']} ({dict_n['id']})"
            blocking = {}
            queue_size = {}
            for io in dict_n['ioInfo']:
                dict_io = io[1]
                port_name = dict_io['name']
                blocking[port_name] = dict_io['blocking']
                queue_size[port_name] = dict_io['queueSize']
            dai_nodes[dict_n['id']]['blocking'] = blocking
            dai_nodes[dict_n['id']]['queue_size'] = queue_size

        print("\nNodes (id):\n===========")
        for id in sorted(dai_nodes):
            print(f"{dai_nodes[id]['name']}")

        # create the nodes.
        qt_nodes = {}
        for id, node in dai_nodes.items():
            qt_nodes[id] = graph.create_node('dai.DepthaiNode', name=node['name'],
                                             color=node_color.get(node['type'], default_node_color),
                                             text_color=(0, 0, 0), push_undo=False)

        nodes = {} # First save all nodes and their inputs/outputs
        for c in dai_connections:
            src_node_id = c["node1Id"]
            src_node = qt_nodes[src_node_id]
            src_port_name = c["node1Output"]
            dst_node_id = c["node2Id"]
            dst_node = qt_nodes[dst_node_id]
            dst_port_name = c["node2Input"]

            if dst_node not in nodes:
                nodes[dst_node] = {'outputs': [], 'inputs': []}
            if src_node not in nodes:
                nodes[src_node] = {'outputs': [], 'inputs': []}

            dst_port_color = (249, 75, 0) if dai_nodes[dst_node_id]['blocking'][dst_port_name] else (0, 255, 0)
            dst_port_label = f"[{dai_nodes[dst_node_id]['queue_size'][dst_port_name]}] {dst_port_name}"

            nodes[dst_node]['inputs'].append({'name': dst_port_name, 'color': dst_port_color, 'label': dst_port_label})
            nodes[src_node]['outputs'].append({'name': src_port_name})

        for node, vals in nodes.items():
            # Go through all inputs/outputs, sort them by name (alphabetical order)
            vals['inputs'] = list(sorted(vals['inputs'], key = lambda el: el['name']))
            vals['outputs'] = list(sorted(vals['outputs'], key=lambda el: el['name']))
            # Create node input/output
            for output in vals['outputs']:
                if not output['name'] in list(node.outputs()):
                    node.add_output(name=output['name'])
            for input in vals['inputs']:
                if not input['name'] in list(node.inputs()):
                    node.add_input(name=input['label'], color=input['color'], multi_input=True)

        print("\nConnections:\n============")
        i = 0
        for c in dai_connections:
            src_node_id = c["node1Id"]
            src_node = qt_nodes[src_node_id]
            src_port_name = c["node1Output"]
            dst_node_id = c["node2Id"]
            dst_node = qt_nodes[dst_node_id]
            dst_port_name = c["node2Input"]
            dst_port_label = f"[{dai_nodes[dst_node_id]['queue_size'][dst_port_name]}] {dst_port_name}"
            # Create the connection between nodes
            print(i,
                  f"{dai_nodes[src_node_id]['name']}: {src_port_name} -> {dai_nodes[dst_node_id]['name']}: {dst_port_label}")
            src_node.outputs()[src_port_name].connect_to(dst_node.inputs()[dst_port_label], push_undo=False)
            i += 1

        # Lock the ports
        graph.lock_all_ports()

        graph_widget.show()
        graph.auto_layout_nodes()
        graph.fit_to_selection()
        graph.set_zoom(-0.9)
        graph.clear_selection()
        graph.clear_undo_stack()
        app.exec_()



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.readlines()

setup(
    name="depthai_pipeline_graph",
    version="0.0.4",
    packages=[*find_packages()],
    entry_points={
        'console_scripts': [
            'pipeline_graph = depthai_pipeline_graph.pipeline_graph:main'
        ]
    },
    author="geaxgx",
    include_package_data=True,
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    description="Tool to create graphs oh DepthAI pipelines",


    python_requires=">=3.6",
    install_requires=list(open("requirements.txt")),
)



#!/usr/bin/python
from Qt import QtWidgets, QtCore, QtGui, QtCompat

from ..custom_widgets.properties import NodePropWidget


class PropertiesDelegate(QtWidgets.QStyledItemDelegate):

    def paint(self, painter, option, index):
        """
        Args:
            painter (QtGui.QPainter):
            option (QtGui.QStyleOptionViewItem):
            index (QtCore.QModelIndex):
        """
        painter.save()
        painter.setRenderHint(QtGui.QPainter.Antialiasing, False)
        painter.setPen(QtCore.Qt.NoPen)

        # draw background.
        bg_clr = option.palette.midlight().color()
        painter.setBrush(QtGui.QBrush(bg_clr))
        painter.drawRect(option.rect)

        # draw border.
        border_width = 1
        if option.state & QtWidgets.QStyle.State_Selected:
            bdr_clr = option.palette.highlight().color()
            painter.setPen(QtGui.QPen(bdr_clr, 1.5))
        else:
            bdr_clr = option.palette.alternateBase().color()
            painter.setPen(QtGui.QPen(bdr_clr, 1))

        painter.setBrush(QtCore.Qt.NoBrush)
        painter.drawRect(QtCore.QRect(
            option.rect.x() + border_width,
            option.rect.y() + border_width,
            option.rect.width() - (border_width * 2),
            option.rect.height() - (border_width * 2))
        )
        painter.restore()


class PropertiesList(QtWidgets.QTableWidget):

    def __init__(self, parent=None):
        super(PropertiesList, self).__init__(parent)
        self.setItemDelegate(PropertiesDelegate())
        self.setColumnCount(1)
        self.setShowGrid(False)
        self.verticalHeader().hide()
        self.horizontalHeader().hide()

        QtCompat.QHeaderView.setSectionResizeMode(
            self.verticalHeader(), QtWidgets.QHeaderView.ResizeToContents)
        QtCompat.QHeaderView.setSectionResizeMode(
            self.horizontalHeader(), 0, QtWidgets.QHeaderView.Stretch)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

    def wheelEvent(self, event):
        delta = event.delta() * 0.2
        self.verticalScrollBar().setValue(
            self.verticalScrollBar().value() - delta
        )


class PropertiesBinWidget(QtWidgets.QWidget):
    """
    The :class:`NodeGraphQt.PropertiesBinWidget` is a list widget for displaying
    and editing a nodes properties.

    .. image:: _images/prop_bin.png
        :width: 950px

    .. code-block:: python
        :linenos:

        from NodeGraphQt import NodeGraph, PropertiesBinWidget

        # create node graph.
        graph = NodeGraph()

        # create properties bin widget.
        properties_bin = PropertiesBinWidget(parent=None, node_graph=graph)
        properties_bin.show()

    Args:
        parent (QtWidgets.QWidget): parent of the new widget.
        node_graph (NodeGraphQt.NodeGraph): node graph.
    """

    #: Signal emitted (node_id, prop_name, prop_value)
    property_changed = QtCore.Signal(str, str, object)

    def __init__(self, parent=None, node_graph=None):
        super(PropertiesBinWidget, self).__init__(parent)
        self.setWindowTitle('Node Properties')
        self._prop_list = PropertiesList()
        # GX
        # self._limit = QtWidgets.QSpinBox()
        # self._limit.setToolTip('Set display nodes limit.')
        # self._limit.setMaximum(10)
        # self._limit.setMinimum(0)
        # self._limit.setValue(1)
        # self._limit.valueChanged.connect(self.__on_limit_changed)
        self.resize(450, 400)

        self._block_signal = False

        self._lock = False
        # self.btn_lock = QtWidgets.QPushButton('lock')
        # self.btn_lock.setToolTip(
        #     'Lock the properties bin prevent nodes from being loaded.')
        # self.btn_lock.clicked.connect(self.lock_bin)

        btn_clr = QtWidgets.QPushButton('Close')
        btn_clr.setToolTip('Close the properties window.')
        btn_clr.clicked.connect(self.close)

        top_layout = QtWidgets.QHBoxLayout()
        top_layout.setSpacing(2)
        # top_layout.addWidget(self._limit)
        top_layout.addStretch(1)
        # top_layout.addWidget(self.btn_lock)
        top_layout.addWidget(btn_clr)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addLayout(top_layout)
        layout.addWidget(self._prop_list, 1)

        # wire up node graph.
        node_graph.add_properties_bin(self)
        node_graph.node_double_clicked.connect(self.add_node)
        node_graph.nodes_deleted.connect(self.__on_nodes_deleted)
        node_graph.property_changed.connect(self.__on_graph_property_changed)

    def __repr__(self):
        return '<{} object at {}>'.format(self.__class__.__name__, hex(id(self)))

    def __on_prop_close(self, node_id):
        items = self._prop_list.findItems(node_id, QtCore.Qt.MatchExactly)
        [self._prop_list.removeRow(i.row()) for i in items]

    def __on_limit_changed(self, value):
        rows = self._prop_list.rowCount()
        if rows > value:
            self._prop_list.removeRow(rows - 1)

    def __on_nodes_deleted(self, nodes):
        """
        Slot function when a node has been deleted.

        Args:
            nodes (list[str]): list of node ids.
        """
        [self.__on_prop_close(n) for n in nodes]

    def __on_graph_property_changed(self, node, prop_name, prop_value):
        """
        Slot function that updates the property bin from the node graph signal.

        Args:
            node (NodeGraphQt.NodeObject):
            prop_name (str): node property name.
            prop_value (object): node property value.
        """
        # print("-- grap> _on_graph_property_changed", node,  prop_name, prop_value )
        properties_widget = self.prop_widget(node)
        if not properties_widget:
            return

        property_window = properties_widget.get_widget(prop_name)

        if property_window and prop_value != property_window.get_value():
            self._block_signal = True
            property_window.set_value(prop_value)
            self._block_signal = False

    def __on_property_widget_changed(self, node_id, prop_name, prop_value):
        """
        Slot function triggered when a property widget value has changed.

        Args:
            node_id (str): node id.
            prop_name (str): node property name.
            prop_value (object): node property value.
        """
        if not self._block_signal:
            self.property_changed.emit(node_id, prop_name, prop_value)

    def limit(self):
        """
        Returns the limit for how many nodes can be loaded into the bin.

        Returns:
            int: node limit.
        """
        # return int(self._limit.value())
        return 1

    def set_limit(self, limit):
        """
        Set limit of nodes to display.

        Args:
            limit (int): node limit.
        """
        self._limit.setValue(limit)

    def add_node(self, node):
        """
        Add node to the properties bin.

        Args:
            node (NodeGraphQt.NodeObject): node object.
        """
        if self.limit() == 0 or self._lock:
            return

        rows = self._prop_list.rowCount()
        if rows >= self.limit():
            self._prop_list.removeRow(rows - 1)

        itm_find = self._prop_list.findItems(node.id, QtCore.Qt.MatchExactly)
        if itm_find:
            self._prop_list.removeRow(itm_find[0].row())

        self._prop_list.insertRow(0)
        prop_widget = NodePropWidget(node=node)
        prop_widget.property_changed.connect(self.__on_property_widget_changed)
        prop_widget.property_closed.connect(self.__on_prop_close)
        self._prop_list.setCellWidget(0, 0, prop_widget)

        item = QtWidgets.QTableWidgetItem(node.id)
        self._prop_list.setItem(0, 0, item)
        self._prop_list.selectRow(0)

    def remove_node(self, node):
        """
        Remove node from the properties bin.

        Args:
            node (str or NodeGraphQt.BaseNode): node id or node object.
        """
        node_id = node if isinstance(node, str) else node.id
        self.__on_prop_close(node_id)

    def lock_bin(self):
        """
        Lock/UnLock the properties bin.
        """
        self._lock = not self._lock
        if self._lock:
            self.btn_lock.setText('UnLock')
        else:
            self.btn_lock.setText('Lock')

    def clear_bin(self):
        """
        Clear the properties bin.
        """
        self._prop_list.setRowCount(0)

    def prop_widget(self, node):
        """
        Returns the node property widget.

        Args:
            node (str or NodeGraphQt.NodeObject): node id or node object.

        Returns:
            NodePropWidget: node property widget.
        """
        node_id = node if isinstance(node, str) else node.id
        itm_find = self._prop_list.findItems(node_id, QtCore.Qt.MatchExactly)
        if itm_find:
            item = itm_find[0]
            return self._prop_list.cellWidget(item.row(), 0)


if __name__ == '__main__':
    import sys
    from NodeGraphQt import BaseNode, NodeGraph
    from NodeGraphQt.constants import (NODE_PROP_QLABEL,
                                       NODE_PROP_QLINEEDIT,
                                       NODE_PROP_QCOMBO,
                                       NODE_PROP_QSPINBOX,
                                       NODE_PROP_COLORPICKER,
                                       NODE_PROP_SLIDER)


    class TestNode(BaseNode):
        NODE_NAME = 'test node'

        def __init__(self):
            super(TestNode, self).__init__()
            self.create_property('label_test', 'foo bar',
                                 widget_type=NODE_PROP_QLABEL)
            self.create_property('text_edit', 'hello',
                                 widget_type=NODE_PROP_QLINEEDIT)
            self.create_property('color_picker', (0, 0, 255),
                                 widget_type=NODE_PROP_COLORPICKER)
            self.create_property('integer', 10,
                                 widget_type=NODE_PROP_QSPINBOX)
            self.create_property('list', 'foo',
                                 items=['foo', 'bar'],
                                 widget_type=NODE_PROP_QCOMBO)
            self.create_property('range', 50,
                                 range=(45, 55),
                                 widget_type=NODE_PROP_SLIDER)

    def prop_changed(node_id, prop_name, prop_value):
        print('-'*100)
        print(node_id, prop_name, prop_value)


    app = QtWidgets.QApplication(sys.argv)

    graph = NodeGraph()
    graph.register_node(TestNode)

    prop_bin = PropertiesBinWidget(node_graph=graph)
    prop_bin.property_changed.connect(prop_changed)

    node = graph.create_node('nodeGraphQt.nodes.TestNode')

    prop_bin.add_node(node)
    prop_bin.show()

    app.exec_()



import sys

import cv2

if len(sys.argv) <= 1:
    raise Exception("Specify the path to the video file (.mp4, .mjpeg, .h265, etc.) like `videoCap.py color.mjpeg`")

cap = cv2.VideoCapture(sys.argv[1])
while cap.isOpened():
    ok, frame = cap.read()
    if not ok: break
    cv2.imshow("Video", frame)
    if cv2.waitKey(1) == ord("q"): break



import json
import os
from dataclasses import replace
from enum import Enum
from typing import List, Tuple, Optional, Union, Any, Dict

import cv2
import depthai as dai
import numpy as np
from depthai import ImgDetection

from depthai_sdk.oak_outputs.normalize_bb import NormalizeBoundingBox
from depthai_sdk.visualize.configs import VisConfig, TextPosition, BboxStyle, StereoColor
from depthai_sdk.visualize.encoder import JSONEncoder
from depthai_sdk.visualize.objects import VisDetections, GenericObject, VisText, VisTrail, VisCircle, VisLine
from depthai_sdk.visualize.visualizer_helper import VisualizerHelper


class Platform(Enum):
    """
    Platform on which the visualizer is running.
    """
    ROBOTHUB = 'robothub'
    PC = 'pc'


class Visualizer(VisualizerHelper):
    # Constants
    IS_INTERACTIVE = 'DISPLAY' in os.environ or os.name == 'nt'

    # Fields
    objects: List[GenericObject]
    config: VisConfig
    _frame_shape: Optional[Tuple[int, ...]]

    def __init__(self, scale: float = None, fps: bool = False):
        self.platform: Platform = self._detect_platform()
        self.objects: List[GenericObject] = []
        self._frame_shape = None

        self.config = VisConfig()

        if fps:
            self.output(show_fps=fps)
        if scale:
            self.output(img_scale=scale)

    def add_object(self, obj: GenericObject) -> 'Visualizer':
        """
        Call `set_config`, `set_frame_shape` and `prepare` for the object and add it to the list of objects.
        Args:
            obj: The object to add.

        Returns:
            self
        """
        obj = obj.set_config(self.config).set_frame_shape(self.frame_shape).prepare()
        self.objects.append(obj)
        return self

    def add_detections(self,
                       detections: List[Union[ImgDetection, dai.Tracklet]],
                       normalizer: NormalizeBoundingBox = None,
                       label_map: List[Tuple[str, Tuple]] = None,
                       spatial_points: List[dai.Point3f] = None,
                       is_spatial=False,
                       ) -> 'Visualizer':
        """
        Add detections to the visualizer.

        Args:
            detections: List of detections.
            normalizer: Normalizer object.
            label_map: List of tuples (label, color).
            spatial_points: List of spatial points. None if not spatial.
            is_spatial: Flag that indicates if the detections are spatial.
        Returns:
            self
        """
        detection_overlay = VisDetections(detections=detections,
                                          normalizer=normalizer,
                                          label_map=label_map,
                                          spatial_points=spatial_points,
                                          is_spatial=is_spatial)
        self.add_object(detection_overlay)
        return self

    def add_text(self,
                 text: str,
                 coords: Tuple[int, int] = None,
                 size: int = None,
                 color: Tuple[int, int, int] = None,
                 thickness: int = None,
                 outline: bool = False,
                 bbox: Union[np.ndarray, Tuple[int, int, int, int]] = None,
                 position: TextPosition = TextPosition.TOP_LEFT,
                 padding: int = 10) -> 'Visualizer':
        """
        Add text to the visualizer.

        Args:
            text: Text.
            coords: Coordinates.
            size: Size of the text.
            color: Color of the text.
            thickness: Thickness of the text.
            outline: Flag that indicates if the text should be outlined.
            bbox: Bounding box.
            position: Position.
            padding: Padding.

        Returns:
            self
        """
        text_overlay = VisText(text=text,
                               coords=coords,
                               size=size,
                               color=color,
                               thickness=thickness,
                               outline=outline,
                               bbox=bbox,
                               position=position,
                               padding=padding)
        self.add_object(text_overlay)
        return self

    def add_trail(self,
                  tracklets: List[dai.Tracklet],
                  label_map: List[Tuple[str, Tuple]]) -> 'Visualizer':
        """
        Add a trail to the visualizer.

        Args:
            tracklets: List of tracklets.
            label_map: List of tuples (label, color).

        Returns:
            self
        """
        trail = VisTrail(tracklets=tracklets,
                         label_map=label_map)
        self.add_object(trail)
        return self

    def add_circle(self,
                   coords: Tuple[int, int],
                   radius: int,
                   color: Tuple[int, int, int] = None,
                   thickness: int = None) -> 'Visualizer':
        """
        Add a circle to the visualizer.

        Args:
            coords: Center of the circle.
            radius: Radius of the circle.
            color: Color of the circle.
            thickness: Thickness of the circle.

        Returns:
            self
        """
        circle = VisCircle(coords=coords,
                           radius=radius,
                           color=color,
                           thickness=thickness)
        self.objects.append(circle)
        return self

    def add_line(self,
                 pt1: Tuple[int, int],
                 pt2: Tuple[int, int],
                 color: Tuple[int, int, int] = None,
                 thickness: int = None) -> 'Visualizer':
        """
        Add a line to the visualizer.

        Args:
            pt1: Start coordinates.
            pt2: End coordinates.
            color: Color of the line.
            thickness: Thickness of the line.

        Returns:
            self
        """
        line = VisLine(pt1=pt1,
                       pt2=pt2,
                       color=color,
                       thickness=thickness)
        self.objects.append(line)
        return self

    def draw(self, frame: np.ndarray) -> Optional[np.ndarray]:
        """
        Draw all objects on the frame if the platform is PC. Otherwise, serialize the objects
        and communicate with the RobotHub application.

        Args:
            frame: The frame to draw on.

        Returns:
            np.ndarray if the platform is PC, None otherwise.
        """
        # Draw overlays
        for obj in self.objects:
            obj.draw(frame)

        # Resize frame if needed
        img_scale = self.config.output.img_scale
        if img_scale:
            if isinstance(img_scale, Tuple):
                frame = cv2.resize(frame, img_scale)
            elif isinstance(img_scale, float) and img_scale != 1.0:
                frame = cv2.resize(frame, dsize=None, fx=img_scale, fy=img_scale)

        self.reset()
        return frame

    def serialize(self) -> str:
        """
        Serialize all contained objects to JSON.

        Returns:
            Stringified JSON.
        """
        parent = {
            'platform': self.platform.value,
            'frame_shape': self.frame_shape,
            'config': self.config,
            'objects': [obj.serialize() for obj in self.objects]
        }

        return json.dumps(parent, cls=JSONEncoder)

    def reset(self):
        """Clear all objects."""
        self.objects.clear()

    def output(self,
               img_scale: float = None,
               show_fps: bool = None) -> 'Visualizer':
        """
        Configure the output of the visualizer.

        Args:
            img_scale: Scale of the output image.
            show_fps: Flag that indicates if the FPS should be shown.

        Returns:
            self
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.output = replace(self.config.output, **kwargs)
        return self

    def stereo(self,
               colorize: StereoColor = None,
               colormap: int = None,
               wls_filter: bool = None,
               wls_lambda: float = None,
               wls_sigma: float = None):
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.stereo = replace(self.config.stereo, **kwargs)
        return self

    def detections(self,
                   thickness: int = None,
                   fill_transparency: float = None,
                   box_roundness: float = None,
                   color: Tuple[int, int, int] = None,
                   bbox_style: BboxStyle = None,
                   line_width: float = None,
                   line_height: float = None,
                   hide_label: bool = None,
                   label_position: TextPosition = None,
                   label_padding: int = None) -> 'Visualizer':
        """
        Configure how bounding boxes will look like.
        Args:
            thickness: Thickness of the bounding box.
            fill_transparency: Transparency of the bounding box.
            box_roundness: Roundness of the bounding box.
            color: Color of the bounding box.
            bbox_style: Style of the bounding box.
            line_width: Width of the bbox horizontal lines CORNERS or ROUNDED_CORNERS style is used.
            line_height: Height of the bbox vertical lines when CORNERS or ROUNDED_CORNERS style is used.
            hide_label: Flag that indicates if the label should be hidden.
            label_position: Position of the label.
            label_padding: Padding of the label.

        Returns:
            self
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.detection = replace(self.config.detection, **kwargs)

        return self

    def text(self,
             font_face: int = None,
             font_color: Tuple[int, int, int] = None,
             font_transparency: float = None,
             font_scale: float = None,
             font_thickness: int = None,
             font_position: TextPosition = None,
             bg_transparency: float = None,
             bg_color: Tuple[int, int, int] = None,
             line_type: int = None,
             auto_scale: bool = None) -> 'Visualizer':
        """
        Configure how text will look like.

        Args:
            font_face: Font face (from cv2).
            font_color: Font color.
            font_transparency: Font transparency.
            font_scale: Font scale.
            font_thickness: Font thickness.
            font_position: Font position.
            bg_transparency: Text background transparency.
            bg_color: Text background color.
            line_type: Line type (from cv2).
            auto_scale: Flag that indicates if the font scale should be automatically adjusted.

        Returns:
            self
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.text = replace(self.config.text, **kwargs)

        return self

    def tracking(self,
                 max_length: int = None,
                 deletion_lost_threshold: int = None,
                 line_thickness: int = None,
                 fading_tails: bool = None,
                 line_color: Tuple[int, int, int] = None,
                 line_type: int = None,
                 bg_color: Tuple[int, int, int] = None) -> 'Visualizer':
        """
        Configure how tracking trails will look like.

        Args:
            max_length: Maximum length of the trail (in pixels).
            deletion_lost_threshold: Number of consequent LOST statuses after which the trail is deleted.
            line_thickness: Thickness of the line.
            fading_tails: Flag that indicates if the tails should fade.
            line_color: Color of the line.
            line_type: Type of the line (from cv2).
            bg_color: Text background color.

        Returns:
            self
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.tracking = replace(self.config.tracking, **kwargs)

        return self

    def _detect_platform(self) -> Platform:
        """
        Detect the platform on which the visualizer is running.

        Returns:
            Platform
        """
        return Platform.PC if self.IS_INTERACTIVE else Platform.ROBOTHUB

    @property
    def frame_shape(self) -> Tuple[int, ...]:
        return self._frame_shape

    @frame_shape.setter
    def frame_shape(self, shape: Tuple[int, ...]) -> None:
        self._frame_shape = shape

    @staticmethod
    def _process_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Process the kwargs and remove all None values."""
        kwargs.pop('self')
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return kwargs



#!/usr/bin/python
from distutils.version import LooseVersion

from Qt import QtGui, QtCore

from ..errors import NodeMenuError
from ..widgets.actions import BaseMenu, GraphAction, NodeAction


class NodeGraphMenu(object):
    """
    The ``NodeGraphMenu`` is the main context menu triggered from the node graph.

    example for accessing the node graph context menu.

    .. code-block:: python
        :linenos:

        from NodeGraphQt import NodeGraph

        node_graph = NodeGraph()

        # get the context menu for the node graph.
        context_menu = node_graph.get_context_menu('graph')

    """

    def __init__(self, graph, qmenu):
        self._graph = graph
        self._qmenu = qmenu

    def __repr__(self):
        return '<{}("{}") object at {}>'.format(
            self.__class__.__name__, self.name(), hex(id(self)))

    @property
    def qmenu(self):
        """
        The underlying qmenu.

        Returns:
            BaseMenu: qmenu object.
        """
        return self._qmenu

    def name(self):
        """
        Returns the name for the menu.

        Returns:
            str: label name.
        """
        return self.qmenu.title()

    def get_menu(self, name):
        """
        Returns the child menu by name.

        Args:
            name (str): name of the menu.

        Returns:
            NodeGraphQt.NodeGraphMenu: menu item.
        """
        menu = self.qmenu.get_menu(name)
        if menu:
            return NodeGraphMenu(self._graph, menu)

    def get_command(self, name):
        """
        Returns the child menu command by name.

        Args:
            name (str): name of the command.

        Returns:
            NodeGraphQt.MenuCommand: context menu command.
        """
        for action in self.qmenu.actions():
            if not action.menu() and action.text() == name:
                return NodeGraphCommand(self._graph, action)

    def all_commands(self):
        """
        Returns all child and sub child commands from the current context menu.

        Returns:
            list[NodeGraphQt.MenuCommand]: list of commands.
        """
        def get_actions(menu):
            actions = []
            for action in menu.actions():
                if not action.menu():
                    if not action.isSeparator():
                        actions.append(action)
                else:
                    actions += get_actions(action.menu())
            return actions
        child_actions = get_actions(self.qmenu)
        return [NodeGraphCommand(self._graph, a) for a in child_actions]

    def add_menu(self, name):
        """
        Adds a child menu to the current menu.

        Args:
            name (str): menu name.

        Returns:
            NodeGraphQt.NodeGraphMenu: the appended menu item.
        """
        menu = BaseMenu(name, self.qmenu)
        self.qmenu.addMenu(menu)
        return NodeGraphMenu(self._graph, menu)

    def add_command(self, name, func=None, shortcut=None):
        """
        Adds a command to the menu.

        Args:
            name (str): command name.
            func (function): command function eg. "func(``graph``)".
            shortcut (str): shotcut key.

        Returns:
            NodeGraphQt.NodeGraphCommand: the appended command.
        """
        action = GraphAction(name, self._graph.viewer())
        action.graph = self._graph
        if LooseVersion(QtCore.qVersion()) >= LooseVersion('5.10'):
            action.setShortcutVisibleInContextMenu(True)
        if shortcut:
            action.setShortcut(shortcut)
        if func:
            action.executed.connect(func)
        qaction = self.qmenu.addAction(action)
        return NodeGraphCommand(self._graph, qaction)

    def add_separator(self):
        """
        Adds a separator to the menu.
        """
        self.qmenu.addSeparator()


class NodesMenu(NodeGraphMenu):
    """
    The ``NodesMenu`` is the context menu triggered from a node.

    **Inherited from:** :class:`NodeGraphQt.NodeGraphMenu`

    example for accessing the nodes context menu.

    .. code-block:: python
        :linenos:

        from NodeGraphQt import NodeGraph

        node_graph = NodeGraph()

        # get the nodes context menu.
        nodes_menu = node_graph.get_context_menu('nodes')
    """

    def add_command(self, name, func=None, node_type=None, node_class=None):
        """
        Re-implemented to add a command to the specified node type menu.

        Args:
            name (str): command name.
            func (function): command function eg. "func(``graph``, ``node``)".
            node_type (str): specified node type for the command.
            node_class (class): specified node class for the command.

        Returns:
            NodeGraphQt.NodeGraphCommand: the appended command.
        """
        if not node_type and not node_class:
            raise NodeMenuError('Node type or Node class not specified!')

        if node_class:
            node_type = node_class.__name__

        node_menu = self.qmenu.get_menu(node_type)
        if not node_menu:
            node_menu = BaseMenu(node_type, self.qmenu)

            if node_class:
                node_menu.node_class = node_class
                node_menu.graph = self._graph

            self.qmenu.addMenu(node_menu)

        if not self.qmenu.isEnabled():
            self.qmenu.setDisabled(False)

        action = NodeAction(name, self._graph.viewer())
        action.graph = self._graph
        if LooseVersion(QtCore.qVersion()) >= LooseVersion('5.10'):
            action.setShortcutVisibleInContextMenu(True)
        if func:
            action.executed.connect(func)

        if node_class:
            node_menus = self.qmenu.get_menus(node_class)
            if node_menu in node_menus:
                node_menus.remove(node_menu)
            for menu in node_menus:
                menu.addAction(action)

        qaction = node_menu.addAction(action)
        return NodeGraphCommand(self._graph, qaction)


class NodeGraphCommand(object):
    """
    Node graph menu command.
    """

    def __init__(self, graph, qaction):
        self._graph = graph
        self._qaction = qaction

    def __repr__(self):
        return '<{}("{}") object at {}>'.format(
            self.__class__.__name__, self.name(), hex(id(self)))

    @property
    def qaction(self):
        """
        The underlying qaction.

        Returns:
            GraphAction: qaction object.
        """
        return self._qaction

    def name(self):
        """
        Returns the name for the menu command.

        Returns:
            str: label name.
        """
        return self.qaction.text()

    def set_shortcut(self, shortcut=None):
        """
        Sets the shortcut key combination for the menu command.

        Args:
            shortcut (str): shortcut key.
        """
        shortcut = shortcut or QtGui.QKeySequence()
        self.qaction.setShortcut(shortcut)

    def run_command(self):
        """
        execute the menu command.
        """
        self.qaction.trigger()



'''
This is a helper class that let's you save depth frames into rosbag (.bag), which can be replayed using RealSense Viewer app.
'''
import os
import time
from pathlib import Path
from typing import List

import depthai as dai
import numpy as np
from rosbags.rosbag1 import Writer
from rosbags.serde import cdr_to_ros1, serialize_cdr
from rosbags.typesys import get_types_from_msg, register_types
from rosbags.typesys.types import builtin_interfaces__msg__Time as Time
from rosbags.typesys.types import diagnostic_msgs__msg__KeyValue as KeyValue
from rosbags.typesys.types import geometry_msgs__msg__Quaternion as Quaternion
from rosbags.typesys.types import geometry_msgs__msg__Transform as Transform
from rosbags.typesys.types import geometry_msgs__msg__Vector3 as Vector3
from rosbags.typesys.types import sensor_msgs__msg__Image as Image
from rosbags.typesys.types import sensor_msgs__msg__RegionOfInterest as Roi
from rosbags.typesys.types import std_msgs__msg__Header as Header
from rosbags.typesys.types import std_msgs__msg__UInt32 as UInt32

from depthai_sdk.recorders.abstract_recorder import Recorder

CAMERA_INFO = """
# This message defines meta information for a camera. It should be in a
# camera namespace on topic "camera_info" and accompanied by up to five
# image topics named:
#
#   image_raw - raw data from the camera driver, possibly Bayer encoded
#   image            - monochrome, distorted
#   image_color      - color, distorted
#   image_rect       - monochrome, rectified
#   image_rect_color - color, rectified
#
# The image_pipeline contains packages (image_proc, stereo_image_proc)
# for producing the four processed image topics from image_raw and
# camera_info. The meaning of the camera parameters are described in
# detail at http://www.ros.org/wiki/image_pipeline/CameraInfo.
#
# The image_geometry package provides a user-friendly interface to
# common operations using this meta information. If you want to, e.g.,
# project a 3d point into image coordinates, we strongly recommend
# using image_geometry.
#
# If the camera is uncalibrated, the matrices D, K, R, P should be left
# zeroed out. In particular, clients may assume that K[0] == 0.0
# indicates an uncalibrated camera.

#######################################################################
#                     Image acquisition info                          #
#######################################################################

# Time of image acquisition, camera coordinate frame ID
Header header    # Header timestamp should be acquisition time of image
                 # Header frame_id should be optical frame of camera
                 # origin of frame should be optical center of camera
                 # +x should point to the right in the image
                 # +y should point down in the image
                 # +z should point into the plane of the image


#######################################################################
#                      Calibration Parameters                         #
#######################################################################
# These are fixed during camera calibration. Their values will be the #
# same in all messages until the camera is recalibrated. Note that    #
# self-calibrating systems may "recalibrate" frequently.              #
#                                                                     #
# The internal parameters can be used to warp a raw (distorted) image #
# to:                                                                 #
#   1. An undistorted image (requires D and K)                        #
#   2. A rectified image (requires D, K, R)                           #
# The projection matrix P projects 3D points into the rectified image.#
#######################################################################

# The image dimensions with which the camera was calibrated. Normally
# this will be the full camera resolution in pixels.
uint32 height
uint32 width

# The distortion model used. Supported models are listed in
# sensor_msgs/distortion_models.h. For most cameras, "plumb_bob" - a
# simple model of radial and tangential distortion - is sufficient.
string distortion_model

# The distortion parameters, size depending on the distortion model.
# For "plumb_bob", the 5 parameters are: (k1, k2, t1, t2, k3).
float64[] D

# Intrinsic camera matrix for the raw (distorted) images.
#     [fx  0 cx]
# K = [ 0 fy cy]
#     [ 0  0  1]
# Projects 3D points in the camera coordinate frame to 2D pixel
# coordinates using the focal lengths (fx, fy) and principal point
# (cx, cy).
float64[9]  K # 3x3 row-major matrix

# Rectification matrix (stereo cameras only)
# A rotation matrix aligning the camera coordinate system to the ideal
# stereo image plane so that epipolar lines in both stereo images are
# parallel.
float64[9]  R # 3x3 row-major matrix

# Projection/camera matrix
#     [fx'  0  cx' Tx]
# P = [ 0  fy' cy' Ty]
#     [ 0   0   1   0]
# By convention, this matrix specifies the intrinsic (camera) matrix
#  of the processed (rectified) image. That is, the left 3x3 portion
#  is the normal camera intrinsic matrix for the rectified image.
# It projects 3D points in the camera coordinate frame to 2D pixel
#  coordinates using the focal lengths (fx', fy') and principal point
#  (cx', cy') - these may differ from the values in K.
# For monocular cameras, Tx = Ty = 0. Normally, monocular cameras will
#  also have R = the identity and P[1:3,1:3] = K.
# For a stereo pair, the fourth column [Tx Ty 0]' is related to the
#  position of the optical center of the second camera in the first
#  camera's frame. We assume Tz = 0 so both cameras are in the same
#  stereo image plane. The first camera always has Tx = Ty = 0. For
#  the right (second) camera of a horizontal stereo pair, Ty = 0 and
#  Tx = -fx' * B, where B is the baseline between the cameras.
# Given a 3D point [X Y Z]', the projection (x, y) of the point onto
#  the rectified image is given by:
#  [u v w]' = P * [X Y Z 1]'
#         x = u / w
#         y = v / w
#  This holds for both images of a stereo pair.
float64[12] P # 3x4 row-major matrix


#######################################################################
#                      Operational Parameters                         #
#######################################################################
# These define the image region actually captured by the camera       #
# driver. Although they affect the geometry of the output image, they #
# may be changed freely without recalibrating the camera.             #
#######################################################################

# Binning refers here to any camera setting which combines rectangular
#  neighborhoods of pixels into larger "super-pixels." It reduces the
#  resolution of the output image to
#  (width / binning_x) x (height / binning_y).
# The default values binning_x = binning_y = 0 is considered the same
#  as binning_x = binning_y = 1 (no subsampling).
uint32 binning_x
uint32 binning_y

# Region of interest (subwindow of full camera resolution), given in
#  full resolution (unbinned) image coordinates. A particular ROI
#  always denotes the same window of pixels on the camera sensor,
#  regardless of binning settings.
# The default setting of roi (all values 0) is considered the same as
#  full resolution (roi.width = width, roi.height = height).
RegionOfInterest roi

================================================================================
MSG: std_msgs/Header

#Two-integer timestamp that is expressed as:
# * stamp.sec: seconds (stamp_secs) since epoch (in Python the variable is called 'secs')
# * stamp.nsec: nanoseconds since stamp_secs (in Python the variable is called 'nsecs')
# time-handling sugar is provided by the client library
time stamp
#Frame this data is associated with
# 0: no frame
# 1: global frame
string frame_id

================================================================================
MSG: sensor_msgs/RegionOfInterest
# This message is used to specify a region of interest within an image.
#
# When used to specify the ROI setting of the camera when the image was
# taken, the height and width fields should either match the height and
# width fields for the associated image; or height = width = 0
# indicates that the full resolution image was captured.

uint32 x_offset  # Leftmost pixel of the ROI
                 # (0 if the ROI includes the left edge of the image)
uint32 y_offset  # Topmost pixel of the ROI
                 # (0 if the ROI includes the top edge of the image)
uint32 height    # Height of ROI
uint32 width     # Width of ROI

# True if a distinct rectified ROI should be calculated from the "raw"
# ROI in this message. Typically this should be False if the full image
# is captured (ROI not used), and True if a subwindow is captured (ROI
# used).
bool do_rectify
"""

STREAM_INFO = """
# This message defines meta information for a stream
# The stream type is expressed in the topic name

uint32 fps        # The nominal streaming rate, defined in Hz
string encoding   # Stream's data format
bool is_recommended # Is this stream recommended by RealSense SDK
"""


class RosbagRecorder(Recorder):
    _closed = False
    _frame_init: List[str]

    def update(self, path: Path, device: dai.Device, _):
        '''
        Args:
            path: Path to the folder where rosbag will be saved
            device: depthai.Device object
        '''
        if path.suffix != '.bag':
            path = path / 'recording.bag'
        if path.exists():
            path.unlink()
        self.path = path

        rgb = False  # TODO: support rgb/mono recording as well
        self._frame_init = []
        self.start_nanos = 0
        self.writer = Writer(self.path)
        # Compression will cause error in RealSense
        # self.writer.set_compression(Writer.CompressionFormat.LZ4)
        self.writer.open()

        # sensor_msgs__msg__CameraInfo won't work, as parameters (eg. D, K, R) are in lowercase (d, k, r), so
        # RealSense Viewer doesn't recognize the msg
        self.dir = os.path.dirname(os.path.realpath(__file__))
        register_types(get_types_from_msg(CAMERA_INFO, 'sensor_msgs/msg/CamInfo'))
        from rosbags.typesys.types import sensor_msgs__msg__CamInfo as CamInfo
        self.CamInfo = CamInfo

        self.depth_conn = self.writer.add_connection('/device_0/sensor_0/Depth_0/image/data', Image.__msgtype__,
                                                     latching=1)
        self.depth_meta_conn = self.writer.add_connection('/device_0/sensor_0/Depth_0/image/metadata',
                                                          KeyValue.__msgtype__, latching=1)

        self.write_uint32('/file_version', 2)
        self.write_keyvalues('/device_0/info', {
            'Name': 'OAK camera',
            'Location': '',
            'Debug Op Code': 0,
            'Advanced Mode': 'YES',
            'Product Id': '0000',
        })
        self.write_streamInfo(depth=True)

        self.write_keyvalues('/device_0/sensor_0/info', {'Name': 'Stereo'})
        self.write_keyvalues('/device_0/sensor_0/property', {
            'Depth Units': '0.001000',
            # 'Serial Number': device.getMxId(),
            # 'Library Version': dai.__version__,
            # 'Exposure': '8000.000000',
            # 'Gain': '16.0',
            # 'Enable Auto Exposure': '0.000000',
            # 'Visual Preset': '2.000000',
            # 'Laser Power': '240.000000',
            # 'Emitter Enabled':'1.000000',
            # 'Frames Queue Size': '16.000000',
            # 'Asic Temperature': '35.000000',
            # 'Error Polling Enabled': '1.000000',
            # 'Projector Temperature': '31.000000',
        })
        self.write_transform('/device_0/sensor_0/Depth_0/tf/0')

        self.calib = device.readCalibration()

        if rgb:
            # Color recording isn't yet possible.
            self.write_keyvalues('/device_0/sensor_1/info', {'Name': 'RGB Camera'})
            self.write_keyvalues('/device_0/sensor_1/property', {
                'Backlight Compensation': '0.000000',
                # 'Brightness': '0.000000',
                # 'Contrast': '50.000000',
                # 'Exposure': '6.000000',
                # 'Gain': '64.000000',
                # 'Gamma': '300.000000',
                # 'Hue': '0.000000',
                # 'Saturation': '64.000000',
                # 'Sharpness': '50.000000',
                # 'White Balance': '4600.000000',
                # 'Enable Auto Exposure': '1.000000',
                # 'Enable Auto White Balance': '1.000000',
                # 'Frames Queue Size': '16.000000',
                # 'Power Line Frequency': '3.000000',
            })
            self.write_transform('/device_0/sensor_1/Color_0/tf/0')

            self.rgb_conn = self.writer.add_connection('/device_0/sensor_1/Color_0/image/data', Image.__msgtype__,
                                                       latching=1)
            self.rgb_meta_conn = self.writer.add_connection('/device_0/sensor_1/Color_0/image/metadata',
                                                            KeyValue.__msgtype__, latching=1)

    def _init_stream(self, frame: dai.ImgFrame):
        resolution = (frame.getWidth(), frame.getHeight())

        if frame.getType() == dai.ImgFrame.Type.RAW16:  # Depth
            self.write_depthInfo('/device_0/sensor_0/Depth_0/info/camera_info', resolution, self.calib)
        else:
            raise NotImplementedError('RosBags currently only support recording of depth!')
        #
        # elif frame.getType() == dai.ImgFrame.Type.RAW8: # Mono Cams
        #     pass
        # else: # Color
        #     fourcc = "I420"
        #     self.write_colorInfo('/device_0/sensor_1/Color_0/info/camera_info', resolution, self.calib)

    def write(self, name: str, imgFrame: dai.ImgFrame):
        if name not in self._frame_init:
            self._init_stream(imgFrame)
            self._frame_init.append(name)

        frame = imgFrame.getCvFrame()
        # First frames
        if self.start_nanos == 0: self.start_nanos = time.time_ns()

        rgb = len(frame.shape) == 3
        if rgb:
            frame = frame[:, :, [2, 1, 0]]
        img = Image(header=self.get_current_header(),
                    height=frame.shape[0],
                    width=frame.shape[1],
                    encoding='rgb8' if rgb else 'mono16',
                    is_bigendian=0,
                    step=frame.shape[1] * 2,
                    data=frame.flatten().view(dtype=np.int8))
        type = Image.__msgtype__

        self._write(self.rgb_conn if rgb else self.depth_conn, type, img)

        self.write_keyvalues(self.rgb_meta_conn if rgb else self.depth_meta_conn, {
            'system_time': "%.6f" % time.time(),
            'timestamp_domain': 'System Time',
            'Time Of Arrival': int(time.time())
        }, connection=True)

    def close(self):
        if self._closed: return
        self._closed = True
        print("ROS .bag saved at: ", str(self.path))
        self.writer.close()

    def _write(self, connection, type, data):
        self.writer.write(connection, time.time_ns() - self.start_nanos, cdr_to_ros1(serialize_cdr(data, type), type))

    def write_streamInfo(self, depth=False, rgb=False):
        # Inspired by https://github.com/IntelRealSense/librealsense/blob/master/third-party/realsense-file/rosbag/msgs/realsense_msgs/StreamInfo.h
        register_types(get_types_from_msg(STREAM_INFO, 'realsense_msgs/msg/StreamInfo'))
        from rosbags.typesys.types import realsense_msgs__msg__StreamInfo as StreamInfo

        if depth:
            streamInfo = StreamInfo(fps=30, encoding="mono16", is_recommended=False)
            c = self.writer.add_connection('/device_0/sensor_0/Depth_0/info', streamInfo.__msgtype__)
            self._write(c, streamInfo.__msgtype__, streamInfo)
        if rgb:
            streamInfo = StreamInfo(fps=30, encoding="rgb8", is_recommended=False)
            c = self.writer.add_connection('/device_0/sensor_1/Color_0/info', streamInfo.__msgtype__)
            self._write(c, streamInfo.__msgtype__, streamInfo)

    def write_keyvalues(self, topicOrConnection, array, connection=False):
        type = KeyValue.__msgtype__
        if not connection:
            c = self.writer.add_connection(topicOrConnection, type, latching=1)
        for name in array:
            self._write(topicOrConnection if connection else c, type, KeyValue(key=name, value=str(array[name])))

    def write_uint32(self, topic, uint32):
        type = UInt32.__msgtype__
        c = self.writer.add_connection(topic, type, latching=1)
        self._write(c, type, UInt32(data=uint32))

    # translation: [x,y,z]
    # rotation: [x,y,z,w]
    # We will use depth alignment to color camera in case we record depth
    def write_transform(self, topic, translation=[0, 0, 0], rotation=[0, 0, 0, 0]):
        type = Transform.__msgtype__
        translation = Vector3(x=translation[0], y=translation[1], z=translation[2])
        rotation = Quaternion(x=rotation[0], y=rotation[1], z=rotation[2], w=rotation[3])
        c = self.writer.add_connection(topic, type, latching=1)
        self._write(c, type, Transform(translation=translation, rotation=rotation))

    def write_depthInfo(self, topic, resolution, calibData):
        # Distortion parameters (k1,k2,t1,t2,k3)
        dist = np.array(calibData.getDistortionCoefficients(dai.CameraBoardSocket.RIGHT))

        # Intrinsic camera matrix
        M_right = np.array(calibData.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, resolution[0], resolution[1]))

        R1 = np.array(calibData.getStereoLeftRectificationRotation())

        # Rectification matrix (stereo cameras only)
        H_right = np.matmul(np.matmul(M_right, R1), np.linalg.inv(M_right))

        # Projection/camera matrix
        lr_extrinsics = np.array(calibData.getCameraExtrinsics(dai.CameraBoardSocket.LEFT, dai.CameraBoardSocket.RIGHT))

        self.write_cameraInfo(topic, resolution, M_right, H_right, lr_extrinsics)

    def write_colorInfo(self, topic, resolution, calibData):
        # Distortion parameters (k1,k2,t1,t2,k3)
        dist = np.array(calibData.getDistortionCoefficients(dai.CameraBoardSocket.RGB))

        # Intrinsic camera matrix
        M_color = np.array(calibData.getCameraIntrinsics(dai.CameraBoardSocket.RGB, resolution[0], resolution[1]))
        self.write_cameraInfo(topic, resolution, M_color, np.zeros((3, 3)), np.zeros((4, 4)))

    def write_cameraInfo(self, topic, resolution, intrinsics, rect, project):
        # print(topic, resolution)
        type = self.CamInfo.__msgtype__
        c = self.writer.add_connection(topic, type, latching=1)
        info = self.CamInfo(header=self.get__default_header(),
                            height=resolution[1],
                            width=resolution[0],
                            distortion_model='Brown Conrady',
                            # D=dist[:5], # Doesn't work
                            D=np.zeros(5),  # Distortion parameters (k1,k2,t1,t2,k3)
                            K=intrinsics.flatten(),  # Intrinsic camera matrix
                            R=rect.flatten(),  # Rectification matrix (stereo cameras only)
                            P=project[:3, :].flatten(),  # Projection/camera matrix
                            binning_x=0,
                            binning_y=0,
                            roi=self.get_default_roi())
        self._write(c, type, info)

    def get__default_header(self):
        t = Time(sec=0, nanosec=0)
        return Header(stamp=t, frame_id='0')

    def get_current_header(self):
        t_str_arr = ("%.9f" % time.time()).split('.')
        t = Time(sec=int(t_str_arr[0]), nanosec=int(t_str_arr[1]))
        return Header(stamp=t, frame_id='0')

    def get_default_roi(self):
        return Roi(x_offset=0, y_offset=0, height=0, width=0, do_rectify=False)



from abc import abstractmethod
from pathlib import Path
from typing import Optional, Callable, List

import depthai as dai

from depthai_sdk import FramePacket
from depthai_sdk.oak_outputs.syncing import SequenceNumSync
from depthai_sdk.oak_outputs.xout import XoutFrames
from depthai_sdk.oak_outputs.xout_base import XoutBase
from depthai_sdk.record import Record
from depthai_sdk.recorders.video_recorder import VideoRecorder
from depthai_sdk.visualize import Visualizer


class BaseConfig:
    @abstractmethod
    def setup(self, pipeline: dai.Pipeline, device, names: List[str]) -> List[XoutBase]:
        raise NotImplementedError()


class OutputConfig(BaseConfig):
    """
    Saves callbacks/visualizers until the device is fully initialized. I'll admit it's not the cleanest solution.
    """
    visualizer: Optional[Visualizer]  # Visualization
    output: Callable  # Output of the component (a callback)
    callback: Callable  # Callback that gets called after syncing

    def __init__(self,
                 output: Callable,
                 callback: Callable,
                 visualizer: Visualizer = None,
                 record_path: Optional[str] = None):
        self.output = output
        self.callback = callback
        self.visualizer = visualizer
        self.record_path = record_path

    def find_new_name(self, name: str, names: List[str]):
        while True:
            arr = name.split(' ')
            num = arr[-1]
            if num.isnumeric():
                arr[-1] = str(int(num) + 1)
                name = " ".join(arr)
            else:
                name = f"{name} 2"
            if name not in names:
                return name

    def setup(self, pipeline: dai.Pipeline, device, names: List[str]) -> List[XoutBase]:
        xoutbase: XoutBase = self.output(pipeline, device)
        xoutbase.setup_base(self.callback)

        if xoutbase.name in names:  # Stream name already exist, append a number to it
            xoutbase.name = self.find_new_name(xoutbase.name, names)
        names.append(xoutbase.name)

        recorder = None
        if self.record_path:
            recorder = VideoRecorder()
            recorder.update(Path(self.record_path), device, [xoutbase])

        if self.visualizer:
            xoutbase.setup_visualize(visualizer=self.visualizer, name=xoutbase.name)

        if self.record_path:
            xoutbase.setup_recorder(recorder=recorder)

        return [xoutbase]


class RecordConfig(BaseConfig):
    rec: Record
    outputs: List[Callable]

    def __init__(self, outputs: List[Callable], rec: Record):
        self.outputs = outputs
        self.rec = rec

    def setup(self, pipeline: dai.Pipeline, device: dai.Device, _) -> List[XoutBase]:
        xouts: List[XoutFrames] = []
        for output in self.outputs:
            xoutbase: XoutFrames = output(pipeline, device)
            xoutbase.setup_base(None)
            xouts.append(xoutbase)

        self.rec.setup_base(None)
        self.rec.start(device, xouts)

        return [self.rec]


class SyncConfig(BaseConfig, SequenceNumSync):
    outputs: List[Callable]
    cb: Callable
    visualizer: Visualizer

    def __init__(self, outputs: List[Callable], callback: Callable):
        self.outputs = outputs
        self.cb = callback

        SequenceNumSync.__init__(self, len(outputs))

        self.packets = dict()

    def new_packet(self, packet: FramePacket, _=None):
        # print('new packet', packet, packet.name, 'seq num',packet.imgFrame.getSequenceNum())
        synced = self.sync(
            packet.imgFrame.getSequenceNum(),
            packet.name,
            packet
        )
        if synced:
            self.cb(synced)

    def setup(self, pipeline: dai.Pipeline, device: dai.Device, _) -> List[XoutBase]:
        xouts = []
        for output in self.outputs:
            xoutbase: XoutBase = output(pipeline, device)
            xoutbase.setup_base(self.new_packet)
            xouts.append(xoutbase)

            xoutbase.setup_visualize(Visualizer(), xoutbase.name)

        return xouts
